"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import json
import base64
import zlib

from .iotcontrol import (
    Label,
    DeviceView, DeviceViewConfig,
    Menu, MenuConfig,
    ButtonGroup, ButtonGroupConfig,
    Button, ButtonConfig,
    TextBox, TextBoxConfig,
    Chart, ChartConfig,
    Dial, DialConfig,
    ColorPicker, ColorPickerConfig,
    TimeGraph, TimeGraphConfig,
    Selector,
    Slider, SliderConfig,
    Direction, DirectionConfig,
    EventLog,
    Knob, KnobConfig,
    AudioVisualDisplay, ControlConfig
)

CONTROL_INSTANCE_DICT = {
    "AVD": AudioVisualDisplay,
    "DVVW": DeviceView,
    "MENU": Menu,
    "BTGP": ButtonGroup,
    "BTTN": Button,
    "TEXT": TextBox,
    "CHRT": Chart,
    "DIAL": Dial,
    "CLR": ColorPicker,
    "TGRPH": TimeGraph,
    "KNOB": Knob,
    "SLCTR": Selector,
    "SLDR": Slider,
    "DIR": Direction,
    "LOG": EventLog,
    "LBL": Label
}

CONFIG_INSTANCE_DICT = {
    "AVD": ControlConfig,
    "DVVW": DeviceViewConfig,
    "MENU": MenuConfig,
    "BTGP": ButtonGroupConfig,
    "BTTN": ButtonConfig,
    "TEXT": TextBoxConfig,
    "CHRT": ChartConfig,
    "DIAL": DialConfig,
    "CLR": ColorPickerConfig,
    "TGRPH": TimeGraphConfig,
    "KNOB": KnobConfig,
    "SLCTR": ControlConfig,
    "SLDR": SliderConfig,
    "DIR": DirectionConfig,
    "LOG": ControlConfig,
    "LBL": ControlConfig
}


def decode_cfg64(cfg: str) -> dict:
    """decodes a CFG from iotdasboard app

    Parameters
    ----------
    cfg : str
        A base64 zipped json

    Returns
    -------
    dict
        A dictionary representing the cfg.
    """
    if not cfg:
        return {}
    ztmp_b = base64.b64decode(cfg)
    tmp_b = zlib.decompress(ztmp_b, wbits=-9)  # Deal with missing header
    try:
        cfg_dict = json.loads(tmp_b)
    except json.JSONDecodeError:
        return ""
    return cfg_dict


def encode_cfg64(cfg: dict) -> str:
    """Encodes cfg dictionary into C64 string

    Args:
        cfg (dict): the config as sent to the IoTDashboard app

    Returns:
        str: the cfg encoded as C64 format
    """
    cfg_json = json.dumps(cfg)
    compress = zlib.compressobj(6, zlib.DEFLATED, -9, zlib.DEF_MEM_LEVEL, zlib.Z_DEFAULT_STRATEGY)
    tmp_z = compress.compress(cfg_json.encode())
    tmp_z += compress.flush()
    return base64.b64encode(tmp_z).decode()


def get_control_from_config(control_type: str, control_id: str, cfg_dict: dict):
    """Returns a control object instantiated from cfg_dict

    Parameters
    ----------
    control_type : str
        The control type of the control within the dict to return
    control_id : str
        The control_id of the control within the dict to instantiate
    cfg_dict : dict
        dictionary of the CFG loaded by decode_cfg from a CFG64 or json

    Returns
    -------
        A control object.
    """
    control_list = cfg_dict[control_type]
    if isinstance(control_list, list):
        for control in control_list:
            if control["controlID"] == control_id:
                return CONTROL_INSTANCE_DICT[control_type].from_cfg_dict(control)
    raise Exception(f"Control with control_id: {control_id} not found")


def get_control_dict_from_config(control_type: str, control_id: str, cfg_dict: dict) -> dict:
    """Returns a control config dictionary from cfg_dict

    Parameters
    ----------
    control_type : str
        The control type of the control within the dict to return
    control_id : str
        The control_id of the control within the dict to return
    cfg_dict : dict
        dictionary of the CFG loaded by decode_cfg from a CFG64 or json

    Returns
    -------
        A dictionary for a control.
    """
    control_list = cfg_dict[control_type]
    if isinstance(control_list, list):
        for control in control_list:
            if control["controlID"] == control_id:
                return control
    raise Exception(f"Control with control_id: {control_id} not found")


def load_all_controls_from_config(device, cfg_dict):
    """Loads all the controls in cfg_dict into device.

    Parameters
    ----------
    device : Dashio.Device
        The device to attach the controls to
    cfg_dict : Dict
        dictionary of the CFG loaded by decode_cfg from a CFG64 or json

    """
    for control_type, control_list in cfg_dict.items():
        if isinstance(control_list, list):
            for control in control_list:
                key = f"{control_type}\t{control['controlID']}"
                if device.is_control_loaded(control_type, control['controlID']):
                    cfg = CONFIG_INSTANCE_DICT[control_type].from_dict(control)
                    device.control_dict[key].add_config_columnar(cfg)
                else:
                    new_control = CONTROL_INSTANCE_DICT[control_type].from_cfg_dict(control)
                    device.add_control(new_control)



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ..constants import BAD_CHARS
from .control import Control, ControlPosition, ControlConfig, _get_title_position, _get_chart_x_axis_labels_style
from .enums import Color, ChartLineType, ChartXAxisLabelsStyle, TitlePosition


class ChartLine:
    """ChartLine class
    """
    def __init__(self, name="", line_type=ChartLineType.LINE, color=Color.BLACK, right_axis=False):
        self.name = name.translate(BAD_CHARS)
        self.line_type = line_type
        self.color = color
        self.data = []
        self.axis_side = 'left'
        if right_axis:
            self.axis_side = 'right'

    def get_line_data(self):
        """Returns the line data formated for the iotdashboard app

        Returns
        -------
        str
            The formatted line data
        """
        data_str = f"\t{self.name}\t{self.line_type.value}\t{self.color.value}\t{self.axis_side}\t"
        data_str += "\t".join(map(str, self.data))
        data_str += "\n"
        return data_str


class ChartConfig(ControlConfig):
    """ChartConfig"""
    def __init__(
        self,
        control_id,
        title: str,
        title_position: TitlePosition,
        x_axis_label: str,
        x_axis_min: float,
        x_axis_max: float,
        x_axis_num_bars: int,
        x_axis_labels_style: ChartXAxisLabelsStyle,
        y_axis_label: str,
        y_axis_min: float,
        y_axis_max: float,
        y_axis_num_bars: int,
        y_axis_label_rt: str,
        y_axis_min_rt: float,
        y_axis_max_rt: float,
        control_position: ControlPosition,
    ) -> None:
        super().__init__(control_id, title, control_position, title_position)
        self.cfg["xAxisLabel"] = x_axis_label.translate(BAD_CHARS)
        self.cfg["xAxisMin"] = x_axis_min
        self.cfg["xAxisMax"] = x_axis_max
        self.cfg["xAxisNumBars"] = x_axis_num_bars
        self.cfg["xAxisLabelsStyle"] = x_axis_labels_style.value
        self.cfg["yAxisLabel"] = y_axis_label.translate(BAD_CHARS)
        self.cfg["yAxisMin"] = y_axis_min
        self.cfg["yAxisMax"] = y_axis_max
        self.cfg["yAxisNumBars"] = y_axis_num_bars
        self.cfg["yAxisLabelRt"] = y_axis_label_rt.translate(BAD_CHARS)
        self.cfg["yAxisMinRt"] = y_axis_min_rt
        self.cfg["yAxisMaxRt"] = y_axis_max_rt

    @classmethod
    def from_dict(cls, cfg_dict: dict):
        """Instantiates GraphConfig from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        GraphConfig
        """
        tmp_cls = cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            _get_title_position(cfg_dict["titlePosition"]),
            cfg_dict["xAxisLabel"],
            cfg_dict["xAxisMin"],
            cfg_dict["xAxisMax"],
            cfg_dict["xAxisNumBars"],
            _get_chart_x_axis_labels_style(cfg_dict["xAxisLabelsStyle"]),
            cfg_dict["yAxisLabel"],
            cfg_dict["yAxisMin"],
            cfg_dict["yAxisMax"],
            cfg_dict["yAxisNumBars"],
            cfg_dict["yAxisLabelRt"],
            cfg_dict["yAxisMinRt"],
            cfg_dict["yAxisMaxRt"],
            ControlPosition(cfg_dict["xPositionRatio"], cfg_dict["yPositionRatio"], cfg_dict["widthRatio"], cfg_dict["heightRatio"])
        )
        tmp_cls.parent_id = cfg_dict["parentID"]
        return tmp_cls


class Chart(Control):
    """A Chart control

    Parameters
    ----------
    Control :
        The control base class
    """

    def get_state(self):
        """Called by iotdashboard"""
        state_str = ""
        for key, line in self.line_dict.items():
            state_str += self._control_hdr_str + key + line.get_line_data()
        return state_str

    def __init__(
        self,
        control_id,
        title="A Chart",
        title_position=TitlePosition.BOTTOM,
        x_axis_label="",
        x_axis_min=0.0,
        x_axis_max=100.0,
        x_axis_num_bars=5,
        x_axis_labels_style=ChartXAxisLabelsStyle.ON,
        y_axis_label="",
        y_axis_min=0.0,
        y_axis_max=1000.0,
        y_axis_num_bars=5,
        y_axis_label_rt="",
        y_axis_min_rt=0.0,
        y_axis_max_rt=1000.0,
        control_position=None,
    ):
        """A Chart Control

        Parameters
        ----------
        control_id : str
            A unique identifier for this control
        title : str, optional
            The title for this control will be displayed on the iotdashboard app, by default "A Chart"
        title_position : TitlePosition, optional
            The position of the title, by default TitlePosition.BOTTOM
        x_axis_label : str, optional
            The label for the X axis, by default ""
        x_axis_min : float, optional
            Min value for the X axis, by default 0.0
        x_axis_max : float, optional
            Max value for the X axis, by default 100.0
        x_axis_num_bars : int, optional
            Number of bars on the X axis, by default 5
        x_axis_labels_style : ChartXAxisLabelsStyle, optional
            The style for the X axis, by default ChartXAxisLabelsStyle.ON
        y_axis_label : str, optional
            The label for the Y axis,, by default ""
        y_axis_min : float, optional
            Min value for the Y axis, by default 0.0
        y_axis_max : float, optional
            Max value for the Y axis, by default 100.0
        y_axis_num_bars : int, optional
            Number of bars on the Y axis, by default 5
        control_position : ControlPosition, optional
            The position of the control on a DeviceView, by default None
        """
        super().__init__("CHRT", control_id)
        self._cfg_columnar.append(
            ChartConfig(
                control_id,
                title,
                title_position,
                x_axis_label,
                x_axis_min,
                x_axis_max,
                x_axis_num_bars,
                x_axis_labels_style,
                y_axis_label,
                y_axis_min,
                y_axis_max,
                y_axis_num_bars,
                y_axis_label_rt,
                y_axis_min_rt,
                y_axis_max_rt,
                control_position
            )
        )
        self.line_dict = {}
        self._x_axis_min = x_axis_min
        self._x_axis_max = x_axis_max
        self._x_axis_num_bars = x_axis_num_bars
        self._y_axis_min = y_axis_min
        self._y_axis_max = y_axis_max
        self._y_axis_num_bars = y_axis_num_bars
        self._y_axis_min_rt = y_axis_min_rt
        self._y_axis_max_rt = y_axis_max_rt

    @property
    def x_axis_min(self):
        """Returns the X axis minimum value"""
        return self._x_axis_min

    @property
    def x_axis_max(self):
        """Returns the X axis maximum value"""
        return self._x_axis_max

    @property
    def x_axis_num_bars(self):
        """Returns the x axis number of bars"""
        return self._x_axis_num_bars

    @property
    def y_axis_min(self):
        """Returns the Y axis minimum value"""
        return self._y_axis_min

    @property
    def y_axis_max(self):
        """Returns the Y axis maximum value"""
        return self._y_axis_max

    @property
    def y_axis_min_rt(self):
        """Returns the Y axis minimum value right side"""
        return self._y_axis_min_rt

    @property
    def y_axis_max_rt(self):
        """Returns the Y axis maximum value right side"""
        return self._y_axis_max_rt

    @property
    def y_axis_num_bars(self):
        """Returns the y axis number of bars"""
        return self.y_axis_num_bars

    @classmethod
    def from_cfg_dict(cls, cfg_dict: dict):
        """Instatiates Chart from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        Chart
        """
        tmp_cls = cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            _get_title_position(cfg_dict["titlePosition"]),
            cfg_dict["xAxisLabel"],
            cfg_dict["xAxisMin"],
            cfg_dict["xAxisMax"],
            cfg_dict["xAxisNumBars"],
            _get_chart_x_axis_labels_style(cfg_dict["xAxisLabelsStyle"]),
            cfg_dict["yAxisLabel"],
            cfg_dict["yAxisMin"],
            cfg_dict["yAxisMax"],
            cfg_dict["yAxisNumBars"],
            cfg_dict.get("yAxisLabelRt", ''),
            cfg_dict.get("yAxisMinRt", 0.0),
            cfg_dict.get("yAxisMaxRt", 1000.0),
            ControlPosition(cfg_dict["xPositionRatio"], cfg_dict["yPositionRatio"], cfg_dict["widthRatio"], cfg_dict["heightRatio"])
        )
        tmp_cls.parent_id = cfg_dict["parentID"]
        return tmp_cls

    def add_line(self, line_id: str, gline: ChartLine):
        """Add a line to the chart

        Parameters
        ----------
        line_id : str
            A unique to this chart line identifier
        gline : GraphLine
            The line to add
        """
        self.line_dict[line_id.translate(BAD_CHARS)] = gline

    def send_chart(self):
        """Sends the chart to any connected iotdashboard app.
        """
        state_str = ""
        for key, line in self.line_dict.items():
            state_str += self._control_hdr_str + key + line.get_line_data()
        self.state_str = state_str



import json
import unittest

from dashio import Direction
from dashio.iotcontrol.enums import Color, DirectionStyle, Precision


class TestDial(unittest.TestCase):

    def _get_cfg_dict(self, cfg_list: list):
        json_str = cfg_list[0].rpartition('\t')[2]
        return json.loads(json_str)

    def test_direction_direction_value(self):
        test_control = Direction("DIRECTIONID")
        test_control.direction_value = 1
        self.assertEqual(test_control.get_state(), '\t{device_id}\tDIR\tDIRECTIONID\t1\n', "Should be 1")

    def test_direction_direction_text(self):
        test_control = Direction("DIRECTIONID")
        test_control.direction_text = 'THAT WAY'
        self.assertEqual(test_control.get_state(), '\t{device_id}\tDIR\tDIRECTIONID\t0\tTHAT WAY\n', "Should be 1")

    def test_direction_pointer_color(self):
        test_control = Direction("DIRECTIONID", pointer_color=Color.BLUE)
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['pointerColor'], '18', "CFG pointerColor should be '18'")

    def test_direction_calibration_angle(self):
        test_control = Direction("DIRECTIONID", calibration_angle=1.2)
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['calAngle'], 1.2, "CFG calAngle should be 1.2")

    def test_direction_style(self):
        test_control = Direction("DIRECTIONID", style=DirectionStyle.NSEW)
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['style'], 'nsew', "CFG style should be 'nsew'")

    def test_direction_units(self):
        test_control = Direction("DIRECTIONID", units='Volts')
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['units'], 'Volts', "CFG units should be 'Volts'")

    def test_direction_precision(self):
        test_control = Direction("DIRECTIONID", precision=Precision.FIVE)
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['precision'], 5, "CFG precision should be 5")


if __name__ == '__main__':
    unittest.main()



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import logging
import socket
import threading
import time

import shortuuid
import zmq
import serial

from .constants import CONNECTION_PUB_URL
from .device import Device


class SerialConnection(threading.Thread):
    """Setups and manages a connection thread to iotdashboard via TCP."""

    def add_device(self, device: Device):
        """Add a device to the connection

        Parameters
        ----------
        device : dashio.Device
            Add a device to the connection.
        """
        if device.device_id not in self.local_device_id_list:
            device.register_connection(self)
            self.local_device_id_list.append(device.device_id)

    def __init__(self, serial_port='/dev/ttyUSB0', baud_rate=38400, context: zmq.Context = None):
        """TCP Connection

        Parameters
        ---------
            serial_port : str, optional
                Serial port to use. Defaults to "/dev/ttyUSB0".
            baud_rate : int, optional
                Baud rate to use. Defaults to 38400.
            context : optional
                ZMQ context. Defaults to None.
        """

        threading.Thread.__init__(self, daemon=True)
        self.context = context or zmq.Context.instance()
        self.zmq_connection_uuid = "SERIAL:" + shortuuid.uuid()
        self.b_connection_id = self.zmq_connection_uuid.encode()

        self.running = True
        self.local_device_id_list = []
        host_name = socket.gethostname()
        host_list = host_name.split(".")
        # rename for .local mDNS advertising
        self.host_name = f"{host_list[0]}.local"
        self.serial_com = serial.Serial(serial_port, baud_rate, timeout=1.0)
        self.serial_com.flush()
        self.start()
        time.sleep(1)

    def close(self):
        """Close the connection."""
        self.running = False

    def run(self):
        tx_zmq_pub = self.context.socket(zmq.PUB)
        tx_zmq_pub.bind(CONNECTION_PUB_URL.format(id=self.zmq_connection_uuid))

        self.rx_zmq_sub = self.context.socket(zmq.SUB)
        # Subscribe on ALL, and my connection
        self.rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "ALL")
        self.rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "SERIAL")
        self.rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, self.zmq_connection_uuid)
        # rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "ANNOUNCE")

        poller = zmq.Poller()
        poller.register(self.rx_zmq_sub, zmq.POLLIN)

        while self.running:
            try:
                socks = dict(poller.poll(50))
            except zmq.error.ContextTerminated:
                break
            if self.rx_zmq_sub in socks:
                try:
                    [_, data] = self.rx_zmq_sub.recv_multipart()
                except ValueError:
                    #  If there aren't two parts continue.
                    continue
                if not data:
                    continue
                logging.debug("SERIAL Tx:\n%s", data.decode().rstrip())
                self.serial_com.write(data)
            if self.serial_com.in_waiting > 0:
                message = self.serial_com.readline()
                if message:
                    try:
                        logging.debug("SERIAL Rx:\n%s", message.decode())
                        tx_zmq_pub.send_multipart([message, self.zmq_connection_uuid])
                    except UnicodeDecodeError:
                        logging.debug("SERIAL DECODE ERROR Rx:\n%s", message.hex())

        self.serial_com.close()
        tx_zmq_pub.close()
        self.rx_zmq_sub.close()



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import datetime

import dateutil.parser

from ..constants import BAD_CHARS
from .control import Control, ControlPosition, ControlConfig, _get_title_position
from .enums import Color, TimeGraphLineType, TitlePosition
from .ring_buffer import RingBuffer
from .event import Event


class DataPoint:
    """
    A time stamped data point for a Time Graph
    """
    def __init__(self, data):
        """
        A time stamped data point for a time series graph.

        Parameters
        ----------
            data: A data point can be an int, float, or boolean, or string representing a number.
        """
        self.timestamp = datetime.datetime.utcnow().replace(microsecond=0, tzinfo=datetime.timezone.utc)
        if isinstance(data, str):
            data = data.translate(BAD_CHARS)
        self.data_point = data

    def __str__(self):
        return f"{self.timestamp.isoformat()},{self.data_point}"


class TimeGraphLine:
    """A TimeGraphLine for a TimeGraph control
    """
    def __init__(
        self,
        name: str = "",
        line_type: TimeGraphLineType = TimeGraphLineType.LINE,
        color: Color = Color.BLACK,
        max_data_points: int = 60,
        break_data: bool = False,
        right_axis: bool = False
    ):
        """A TimeGraphLine for a TimeGraph control

        Parameters
        ----------
        name : str, optional
            The name of the line, by default ""
        line_type : TimeGraphLineType, optional
            Which line type to use, by default TimeGraphLineType.LINE
        color : Color, optional
            The color of the line, by default Color.BLACK
        max_data_points : int, optional
            The data is stored in a ring buffer this sets the ring buffer size, by default 60
        break_data : bool, optional
            If true draws the line with breaks between missing data, by default False
        """
        self.name = name.translate(BAD_CHARS)
        self.line_type = line_type
        self.color = color
        self.break_data = break_data
        self.data = RingBuffer(max_data_points)
        self.axis_side = 'left'
        if right_axis:
            self.axis_side = 'right'

    def get_line_format(self):
        return f"\t{self.name}\t{self.line_type.value}\t{self.color.value}\t{self.axis_side}\n"

    def get_line_from_timestamp(self, timestamp: str) -> str:
        """Converts data from timestamp to a string formatted for the iotdashboard app

        Returns
        -------
        str
            Formatted line data
        """
        data_str = f"\t{self.name}\t{self.line_type.value}\t{self.color.value}\t{self.axis_side}"

        try:
            d_stamp = dateutil.parser.isoparse(timestamp)
        except ValueError:
            return ""
        first = True
        valid_data = False
        data_list = self.data.get()
        for data_p in data_list:
            if data_p.timestamp > d_stamp:
                if first and self.break_data:
                    data_str += "\t" + f"{data_p.timestamp.isoformat()},b"
                data_str += "\t" + str(data_p)
                valid_data = True
            first = False
        data_str += "\n"
        if not valid_data:
            data_str = ""
        return data_str

    def add_data_point(self, data_point: DataPoint):
        """Add and sends a single datapoint to the line. It automatically timestamps to the current time.

        Parameters
        ----------
            data_point : DataPoint
                A single data point
        """
        data_p = DataPoint(data_point)
        self.data.append(data_p)

    def get_latest_data(self) -> str:
        """Get the last inserted datapoint

        Returns
        -------
        str
            The last datapoint
        """
        if self.data.empty():
            return ""
        data_str = "\t" + str(self.data.get_latest()) + "\n"
        return data_str


class TimeGraphConfig(ControlConfig):
    """TimeGraphConfig"""
    def __init__(
        self,
        control_id: str,
        title: str,
        title_position: TitlePosition,
        y_axis_label: str,
        y_axis_min: float,
        y_axis_max: float,
        y_axis_num_bars: int,
        y_axis_label_rt: str,
        y_axis_min_rt: float,
        y_axis_max_rt: float,
        control_position: ControlPosition
    ) -> None:
        super().__init__(control_id, title, control_position, title_position)
        self.cfg["yAxisLabel"] = y_axis_label.translate(BAD_CHARS)
        self.cfg["yAxisMin"] = y_axis_min
        self.cfg["yAxisMax"] = y_axis_max
        self.cfg["yAxisNumBars"] = y_axis_num_bars
        self.cfg["yAxisLabelRt"] = y_axis_label_rt.translate(BAD_CHARS)
        self.cfg["yAxisMinRt"] = y_axis_min_rt
        self.cfg["yAxisMaxRt"] = y_axis_max_rt

    @classmethod
    def from_dict(cls, cfg_dict: dict):
        """Instantiates TimeGraphConfig from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        SliderConfig
        """
        tmp_cls = cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            _get_title_position(cfg_dict["titlePosition"]),
            cfg_dict["yAxisLabel"],
            cfg_dict["yAxisMin"],
            cfg_dict["yAxisMax"],
            cfg_dict["yAxisNumBars"],
            cfg_dict["yAxisLabelRt"],
            cfg_dict["yAxisMinRt"],
            cfg_dict["yAxisMaxRt"],
            ControlPosition(cfg_dict["xPositionRatio"], cfg_dict["yPositionRatio"], cfg_dict["widthRatio"], cfg_dict["heightRatio"])
        )
        tmp_cls.parent_id = cfg_dict["parentID"]
        return tmp_cls


class TimeGraph(Control):
    """A TimeGraph control
    """

    def __init__(
        self,
        control_id: str,
        title="A TimeGraph",
        title_position=TitlePosition.BOTTOM,
        y_axis_label="Y axis",
        y_axis_min=0.0,
        y_axis_max=1000.0,
        y_axis_num_bars=5,
        y_axis_label_rt='',
        y_axis_min_rt=0.0,
        y_axis_max_rt=1000.0,
        control_position=None,
    ):
        """A Timegraph control

        Parameters
        ----------
        control_id : str
            An unique control identity string. The control identity string must be a unique string for each control per device
        title : str, optional
            Title of the control, by default "A TimeGraph"
        control_position : ControlPosition, optional
            The position of the control on a DeviceView, by default None
        title_position : TitlePosition, optional
            Position of the title when displayed on the iotdashboard app, by default None
        y_axis_label : str, optional
            Label for the Y axis, by default "Y axis"
        y_axis_min : float, optional
            Min value for the Y axis, by default 0.0
        y_axis_max : float, optional
            Max value for the Y axis, by default 100.0
        y_axis_num_bars : int, optional
            The number of bars to display on the graph, by default 5
        """
        super().__init__("TGRPH", control_id)
        self._cfg_columnar.append(
            TimeGraphConfig(
                control_id,
                title,
                title_position,
                y_axis_label,
                y_axis_min,
                y_axis_max,
                y_axis_num_bars,
                y_axis_label_rt,
                y_axis_min_rt,
                y_axis_max_rt,
                control_position
            )
        )
        self._message_rx_event = Event()
        self._message_rx_event += self._get_lines_from_timestamp
        self._y_axis_min = y_axis_min
        self._y_axis_max = y_axis_max
        self._y_axis_num_bars = y_axis_num_bars
        self._y_axis_min_rt = y_axis_min_rt
        self._y_axis_max_rt = y_axis_max_rt

        self.line_dict = {}

    @property
    def y_axis_min(self):
        """Returns the Y axis minimum value"""
        return self._y_axis_min

    @property
    def y_axis_max(self):
        """Returns the Y axis maximum value"""
        return self._y_axis_max

    @property
    def y_axis_min_rt(self):
        """Returns the Y axis minimum value"""
        return self._y_axis_min_rt

    @property
    def y_axis_max_rt(self):
        """Returns the Y axis maximum value"""
        return self._y_axis_max_rt

    @property
    def y_axis_num_bars(self):
        """Returns the y axis number of bars"""
        return self.y_axis_num_bars

    @classmethod
    def from_cfg_dict(cls, cfg_dict: dict):
        """Instatiates TimeGraph from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        TimeGraph
        """
        tmp_cls = cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            _get_title_position(cfg_dict["titlePosition"]),
            cfg_dict["yAxisLabel"],
            cfg_dict["yAxisMin"],
            cfg_dict["yAxisMax"],
            cfg_dict["yAxisNumBars"],
            cfg_dict.get("yAxisLabelRt", ''),
            cfg_dict.get("yAxisMinRt", 0.0),
            cfg_dict.get("yAxisMaxRt", 1000.0),
            ControlPosition(cfg_dict["xPositionRatio"], cfg_dict["yPositionRatio"], cfg_dict["widthRatio"], cfg_dict["heightRatio"])
        )
        tmp_cls.parent_id = cfg_dict["parentID"]
        return tmp_cls

    def get_state(self):
        state_str = ""
        for key, line in self.line_dict.items():
            line_format = line.get_line_format()
            if line_format:
                state_str += self._control_hdr_str + "BRDCST\t" + key + line_format
        return state_str

    def add_line(self, line_id: str, gline: TimeGraphLine):
        """Add a line to the TimeGraph

        Parameters
        ----------
        line_id : str
            A unique to the graph line id
        gline : TimeGraphLine
            The Line to add
        """
        self.line_dict[line_id] = gline

    def _get_lines_from_timestamp(self, msg):
        state_str = ""
        try:
            dashboard_id = msg[3]
            from_timestamp = msg[4]
        except IndexError:
            return ""
        for key, line in self.line_dict.items():
            if line.data:
                line_data = line.get_line_from_timestamp(from_timestamp)
                if line_data:
                    state_str += self._control_hdr_str + dashboard_id + "\t" + key + line_data
        return state_str

    def send_data(self):
        """Sends the latest Data to the iotdashboard app.
        """
        state_str = ""
        for key, line in self.line_dict.items():
            if line.data:
                line_data = line.get_latest_data()
                if line_data:
                    state_str += self._control_hdr_str + key + line_data
        self.state_str = state_str



import json
import unittest

from dashio import Menu
from dashio.iotcontrol.enums import Icon


def _get_cfg_dict(cfg_list: list):
    json_str = cfg_list[0].rpartition('\t')[2]
    return json.loads(json_str)


class TestMenu(unittest.TestCase):

    def test_menu_control_type(self):
        test_control = Menu("MENUID")
        test_str_list = test_control._control_hdr_str.split('\t')
        self.assertEqual(test_str_list[2], 'MENU', "control type should be KNOB")

    def test_menu_control_id(self):
        test_control = Menu("MENUID")
        test_str_list = test_control._control_hdr_str.split('\t')
        self.assertEqual(test_str_list[3], 'MENUID', "control type should be KNOBID")

    def test_menu_cfg_icon_name(self):
        test_control = Menu("MENUID", icon_name=Icon.ON_OFF)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(Icon(cfg_dict['iconName']), Icon.ON_OFF, "CFG iconName should be iconName")

    def test_menu_cfg_text(self):
        test_control = Menu("MENUID", text="TEXT")
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['text'], "TEXT", "CFG text should be TEXT")


if __name__ == '__main__':
    unittest.main()



"""Memory Class"""
import logging
import threading
import zmq

from ..constants import CONNECTION_PUB_URL, TASK_PULL
from .action_station_service_config import (ActionServiceCFG,
                                            BoolParameterSpec,
                                            FloatParameterSpec,
                                            IntParameterSpec,
                                            ListParameterSpec,
                                            SelectorParameterSpec,
                                            SliderParameterSpec,
                                            StringParameterSpec)


def make_test_config(num_tests):
    """Make a timer config"""
    provisioning_list = [
        SelectorParameterSpec("A Selector", ["Selection 1", "Selection2", "Selection3"], "Selection 1"),
        IntParameterSpec("An Int", 0, 100, "jiggers", 42),
        StringParameterSpec("A String", "Little Bo Peep...."),
        FloatParameterSpec("A Float", 2.71828, 299792458.0, "jiffies", 1.4142),
        BoolParameterSpec("A bool", True),
        SliderParameterSpec("A Slider", 1.0, 10.0, 1.0, 5.0),
        ListParameterSpec(
            "List of Things",
            "Create a list",
            [
                IntParameterSpec(
                    "Add a uint_16",
                    0,
                    65535,
                    "",
                    0
                ),
                SelectorParameterSpec(
                    "Choose a thing", [
                        "Thing One",
                        "Thing Two",
                        "The Lorax",
                        "The Grinch",
                        "Horton",
                        "Sam-I-Am",
                        "Thidwick",
                        "Yertle the Turtle",
                        "Lord Droon"
                    ],
                    "Horton",
                )
            ]
        )

    ]
    parameter_list = []

    timer_cfg = ActionServiceCFG(
        "TST",
        "Test",
        "Test with all provisioning types",
        "TST",
        num_tests,
        True,
        True,
        provisioning_list,
        parameter_list
        #  parameter_list_out
    )
    return timer_cfg.__json__()


class ASService(threading.Thread):
    """Action Station Template Class"""

    def send_message(self, out_message=""):
        """Send the message"""
        self.task_sender.send_multipart([b"ALL", out_message.encode('utf-8')])

    def close(self):
        """Close the thread"""
        self.running = False

    def __init__(self, device_id: str, action_station_id: str, control_config_dict: dict, context: zmq.Context) -> None:
        threading.Thread.__init__(self, daemon=True)

        self.context = context
        self.running = True

        self.control_id = control_config_dict['controlID']
        self.name = control_config_dict['name']
        self.control_type = control_config_dict['objectType']
        # provision_list = control_config_dict['provisioning']

        self.sub_url = CONNECTION_PUB_URL.format(id=action_station_id)

        self.push_url = TASK_PULL.format(id=action_station_id)
        self.task_sender = self.context.socket(zmq.PUSH)
        self.task_sender.connect(self.push_url)

        self.control_msg = f"\t{device_id}\t{self.control_type}\t{self.control_id}"
        logging.debug("Init Class: %s, %s", self.control_type, self.name)

        self.start()

    def run(self):
        receiver = self.context.socket(zmq.SUB)
        receiver.connect(self.sub_url)
        receiver.setsockopt_string(zmq.SUBSCRIBE, self.control_msg)
        poller = zmq.Poller()
        poller.register(receiver, zmq.POLLIN)

        while self.running:
            try:
                socks = dict(poller.poll(15))
            except zmq.error.ContextTerminated:
                break
            if receiver in socks:
                message, _ = receiver.recv()
                if message:
                    logging.debug("%s\t%s RX:\n%s", self.control_type, self.control_id, message.decode())

        self.task_sender.close()
        receiver.close()



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import json
import logging
import ssl
import threading
import time

import paho.mqtt.client as mqtt
import shortuuid
import zmq

from .constants import CONNECTION_PUB_URL


class MQTTConnection(threading.Thread):
    """Setups and manages a connection thread to the MQTT Server."""

    def _on_connect(self, client, userdata, flags, msg):
        if msg == 0:
            self._connected = True
            self._disconnected = False
            for device_id in self._device_id_list:
                control_topic = f"{self.username}/{device_id}/control"
                self.mqttc.subscribe(control_topic, 0)
            for device_id in self._device_id_rx_list:
                data_topic = f"{self.username}/{device_id}/data"
                self.mqttc.subscribe(data_topic, 0)
            self._send_dash_announce()
            logging.debug("connected OK")
        else:
            logging.debug("Bad connection Returned code=%s", msg)

    def _on_disconnect(self, client, userdata, msg):
        logging.debug("disconnecting reason  %s", msg)
        self._connected = False
        self._disconnected = True

    def _on_message(self, client, obj, msg):
        data = str(msg.payload, "utf-8").strip()
        logging.debug("MQTT RX:\n%s", data)
        self.tx_zmq_pub.send_multipart([msg.payload, self.b_connection_id])

    def _on_subscribe(self, client, obj, mid, granted_qos):
        logging.debug("Subscribed: %s %s", str(mid), str(granted_qos))

    def _on_log(self, client, obj, level, string):
        logging.debug(string)

    def add_device(self, device):
        """Add a Device to the connextion

        Parameters
        ----------
            device (Device):
                The Device to add.
        """
        if device.device_id not in self._device_id_list:
            self._device_id_list.append(device.device_id)
            device.register_connection(self)
            if self._connected:
                control_topic = f"{self.username}/{device.device_id}/control"
                self.mqttc.subscribe(control_topic, 0)
                self._send_dash_announce()

    def _add_device_rx(self, msg_dict):
        """Connect to another device"""
        device_id = msg_dict["deviceID"]
        logging.debug("MQTT DEVICE CONNECT: %s", device_id)
        if device_id not in self._device_id_rx_list:
            self._device_id_rx_list.append(device_id)
            data_topic = f"{self.username}/{device_id}/data"
            self.mqttc.subscribe(data_topic, 0)

    def _del_device_rx(self, msg_dict):
        device_id = msg_dict["deviceID"]
        if device_id in self._device_id_rx_list:
            data_topic = f"{self.username}/{device_id}/data"
            self.mqttc.unsubscribe(data_topic)
            logging.debug("MQTT DEVICE_DISCONNECT: %s", device_id)
            del self._device_id_rx_list[device_id]

    def _send_dash_announce(self):
        msg = {
            'msgType': 'send_announce',
            'connectionUUID': self.zmq_connection_uuid
        }
        logging.debug("MQTT SEND ANNOUNCE: %s", msg)
        self.tx_zmq_pub.send_multipart([b"COMMAND", json.dumps(msg).encode()])

    def close(self):
        """Close the connection."""
        self.running = False

    def __init__(self, host, port, username="", password="", use_ssl=False, context: zmq.Context = None):
        """
        Setups and manages a connection thread to the MQTT Server.

        Parameters
        ---------
            host : str
                The server name of the mqtt host.
            port :int
                Port number to connect to.
            username : str
                username for the mqtt connection.
            password : str
                password for the mqtt connection.

        Keyword Parameters
        -----------------
            use_ssl : bool
                Whether to use ssl for the connection or not. (default: {False})
        """

        threading.Thread.__init__(self, daemon=True)

        self.context = context or zmq.Context.instance()
        self.zmq_connection_uuid = "MQTT:" + shortuuid.uuid()
        self.b_connection_id = self.zmq_connection_uuid.encode('utf-8')

        self._connected = False
        self._disconnected = True
        self.connection_topic_list = []
        self._device_id_list = []
        self._device_id_rx_list = []
        self.host = host
        self.port = port
        # self.last_will = "OFFLINE"
        self.running = True
        self.username = username
        self.mqttc = mqtt.Client()
        self.disconnect_timeout = 15.0
        # Assign event callbacks
        self.mqttc.on_message = self._on_message
        self.mqttc.on_connect = self._on_connect
        self.mqttc.on_disconnect = self._on_disconnect
        self.mqttc.on_subscribe = self._on_subscribe

        if use_ssl:
            self.mqttc.tls_set(
                ca_certs=None,
                certfile=None,
                keyfile=None,
                cert_reqs=ssl.CERT_REQUIRED,
                tls_version=ssl.PROTOCOL_TLSv1_2,
                ciphers=None,
            )
            self.mqttc.tls_insecure_set(False)

        self.mqttc.on_log = self._on_log
        # self.mqttc.will_set(self.data_topic, self.last_will, qos=1, retain=False)
        # Connect
        if username and password:
            self.mqttc.username_pw_set(username, password)
        try:
            self.mqttc.connect(self.host, self.port)
        except mqtt.socket.gaierror as error:
            logging.debug("No connection to internet: %s", str(error))
        # Start subscribe, with QoS level 0
        self.disconnect_timeout = 15.0
        self.start()

    def _mqtt_command(self, msg_dict: dict):
        logging.debug("MQTT CMD: %s", msg_dict)
        if msg_dict['msgType'] == 'connect':
            self._add_device_rx(msg_dict)
        if msg_dict['msgType'] == 'disconnect':
            self._del_device_rx(msg_dict)

    def run(self):
        self.mqttc.loop_start()

        self.tx_zmq_pub = self.context.socket(zmq.PUB)
        self.tx_zmq_pub.bind(CONNECTION_PUB_URL.format(id=self.zmq_connection_uuid))

        self.rx_zmq_sub = self.context.socket(zmq.SUB)
        self.tx_zmq_pub = self.context.socket(zmq.PUB)

        #  Subscribe on ALL, and my connection
        self.rx_zmq_sub.setsockopt(zmq.SUBSCRIBE, b"ALL")
        self.rx_zmq_sub.setsockopt(zmq.SUBSCRIBE, b"MQTT")
        self.rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, self.zmq_connection_uuid)

        poller = zmq.Poller()
        poller.register(self.rx_zmq_sub, zmq.POLLIN)

        while self.running:
            try:
                socks = dict(poller.poll(50))
            except zmq.error.ContextTerminated:
                break
            if self.rx_zmq_sub in socks:
                try:
                    [msg_to, data] = self.rx_zmq_sub.recv_multipart()
                except ValueError:
                    logging.debug("MQTT value error")
                    continue
                if not data:
                    logging.debug("MQTT no data error")
                    continue
                # logging.debug("DASH: %s ,%s", msg_to, data)
                if msg_to == b'COMMAND':
                    logging.debug("MQTT RX COMMAND")
                    self._mqtt_command(json.loads(data))
                    continue
                msg_l = data.split(b'\t')
                try:
                    device_id = msg_l[1].decode().strip()
                except IndexError:
                    continue
                data_topic = f"{self.username}/{device_id}/data"
                if self._connected:
                    logging.debug("MQTT TX:\n%s", data.decode().rstrip())
                    self.mqttc.publish(data_topic, data.decode())
            if self._disconnected:
                self.disconnect_timeout = min(self.disconnect_timeout, 900)
                time.sleep(self.disconnect_timeout)
                try:
                    self.mqttc.connect(self.host, self.port)
                except mqtt.socket.gaierror as error:
                    logging.debug("No connection to internet: %s", str(error))
                self.disconnect_timeout = self.disconnect_timeout * 2

        self.mqttc.loop_stop()
        self.tx_zmq_pub.close()
        self.rx_zmq_sub.close()



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import json
import socket
import threading
import zmq
import logging
from zeroconf import ServiceBrowser, ServiceInfo, Zeroconf


class ZeroConfDashTCPListener:
    """A zeroc conf listener"""
    def __init__(self, service_type: str, connection_uuid: str, context: zmq.Context):
        self.context = context
        self.service_type = service_type
        self.connection_uuid = connection_uuid
        self.zmq_socket = self.context.socket(zmq.PUSH)
        self.zmq_socket.connect("inproc://zconf")

    def _send_msg(self, msg: dict):
        """Send a message"""
        logging.debug("ZCONF: %s", json.dumps(msg))
        self.zmq_socket.send(json.dumps(msg).encode())

    def _send_info(self, connection_uuid, info):
        try:
            device_ids = info.properties[b'deviceID'].decode()
        except KeyError:
            device_ids = ''
        for address in info.addresses:
            msg = {
                'objectType': 'zeroConfUpdate',
                'address': socket.inet_ntoa(address),
                'deviceID': device_ids,
                'connectionID': connection_uuid,
                'port': str(info.port)
            }
            self._send_msg(msg)

    def remove_service(self, zeroconf, service_type, name):
        """Remove service"""
        connection_uuid = name.split("._", 1)[0]
        if service_type == self.service_type and connection_uuid != self.connection_uuid:
            msg = {
                'objectType': 'zeroConfDisconnect',
                'connectionID': connection_uuid
            }
            self._send_msg(msg)

    def add_service(self, zeroconf, service_type, name):
        """add service"""
        connection_uuid = name.split(".", 1)[0]
        if service_type == self.service_type and connection_uuid != self.connection_uuid:
            info = zeroconf.get_service_info(service_type, name)
            if info:
                self._send_info(connection_uuid, info)

    def update_service(self, zeroconf, service_type, name):
        """update service"""
        connection_uuid = name.split(".", 1)[0]
        if service_type == self.service_type and connection_uuid != self.connection_uuid:
            info = zeroconf.get_service_info(service_type, name)
            if info:
                self._send_info(connection_uuid, info)


class ZeroconfService(threading.Thread):

    def _get_ext_ip_address(self):
        """Find the external IP address."""
        test_s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # doesn't even have to be reachable
            test_s.connect(('10.255.255.255', 1))
            i_address = test_s.getsockname()[0]
        except socket.error:
            i_address = '127.0.0.1'
        finally:
            test_s.close()
        return i_address

    def _zconf_update_zmq(self):
        zconf_desc = {
            'connectionUUID': self.connection_uuid,
            'deviceID': ','.join(self.device_id_list)
        }
        zconf_info = ServiceInfo(
            self.fully_qualified_name,
            f"{self.connection_uuid}._DashIO._tcp.local.",
            addresses=[socket.inet_aton(self.local_ip_address)],
            port=self.local_port,
            properties=zconf_desc,
            server=self.host_name + ".",
        )
        self.zeroconf.update_service(zconf_info)

    def add_device(self, device_id):
        """Add a device_id to the advertiser"""
        if device_id not in self.device_id_list:
            self.device_id_list.append(device_id)
            self._zconf_update_zmq()

    def remove_device_id(self, device_id):
        """Remove a device_id from the advertiser"""
        if device_id in self.device_id_list:
            self.device_id_list.remove(device_id)
            self._zconf_update_zmq()

    def close(self):
        self.zeroconf.remove_all_service_listeners()
        self.zeroconf.unregister_all_services()
        self.zeroconf.close()

    def __init__(self, connection_uuid: str, ip_address: str, port: int, context: zmq.Context = None):
        threading.Thread.__init__(self, daemon=True)
        self.context = context or zmq.Context.instance()
        self.connection_uuid = connection_uuid
        self.fully_qualified_name = "_DashIO._tcp.local."
        host_name = socket.gethostname()
        host_list = host_name.split(".")
        # rename for .local mDNS advertising
        self.host_name = f"{host_list[0]}.local"
        self.local_ip_address = ip_address
        self.local_port = port
        self.zeroconf = Zeroconf()
        self.listener = ZeroConfDashTCPListener(self.fully_qualified_name, self.connection_uuid, self.context)
        self.device_id_list = []
        self.start()

    def run(self):
        self.browser = ServiceBrowser(self.zeroconf, self.fully_qualified_name, self.listener)
        zconf_desc = {
            'connectionUUID': self.connection_uuid,
            'deviceID': ','.join(self.device_id_list)
        }
        zconf_info = ServiceInfo(
            self.fully_qualified_name,
            f"{self.connection_uuid}._DashIO._tcp.local.",
            addresses=[socket.inet_aton(self.local_ip_address)],
            port=self.local_port,
            properties=zconf_desc,
            server=self.host_name + ".",
        )
        self.zeroconf.update_service(zconf_info)



import json
import unittest

from dashio import TimeGraph


def _get_cfg_dict(cfg_list: list):
    json_str = cfg_list[0].rpartition('\t')[2]
    return json.loads(json_str)


class TestTimeGraph(unittest.TestCase):

    def test_time_graph_control_type(self):
        test_control = TimeGraph("TIMEGRAPHID")
        test_str_list = test_control._control_hdr_str.split('\t')
        self.assertEqual(test_str_list[2], 'TGRPH', "control type should be TEXT")

    def test_time_graph_control_id(self):
        test_control = TimeGraph("TIMEGRAPHID")
        test_str_list = test_control._control_hdr_str.split('\t')
        self.assertEqual(test_str_list[3], 'TIMEGRAPHID', "control type should be TEXTBOXID")

    def test_time_graph_cfg_y_axis_label(self):
        test_control = TimeGraph("TIMEGRAPHID", y_axis_label='YAXIS')
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['yAxisLabel'], 'YAXIS', "CFG yAxisLabel should be YAXIS")

    def test_time_graph_cfg_y_axis_min(self):
        test_control = TimeGraph("TIMEGRAPHID", y_axis_min=1.2)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['yAxisMin'], 1.2, "CFG yAxisMin should be 1.2")

    def test_time_graph_cfg_y_axis_max(self):
        test_control = TimeGraph("TIMEGRAPHID", y_axis_max=1.9)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['yAxisMax'], 1.9, "CFG yAxisMin should be 1.9")

    def test_time_graph_cfg_y_axis_num_bars(self):
        test_control = TimeGraph("TIMEGRAPHID", y_axis_num_bars=3)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['yAxisNumBars'], 3, "CFG yAxisNumBars should be 3")


if __name__ == '__main__':
    unittest.main()



import json
import unittest

from dashio import DeviceView
from dashio.iotcontrol.enums import Color, Icon


def _get_cfg_dict(cfg_list: list):
    json_str = cfg_list[0].rpartition('\t')[2]
    return json.loads(json_str)


class TestDeviceView(unittest.TestCase):

    def test_device_view_cfg_icon_name(self):
        test_control = DeviceView("DVIEWID", icon=Icon.RIGHT_DOUBLE)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(Icon(cfg_dict['iconName']), Icon.RIGHT_DOUBLE, "CFG iconName should be RIGHT_DOUBLE")

    def test_device_view_cfg_color(self):
        test_control = DeviceView("DVIEWID", color=Color.BURLY_WOOD)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(Color(int(cfg_dict['color'])), Color.BURLY_WOOD, "CFG color should be BURLY_WOOD")

    def test_device_view_cfg_share_column(self):
        test_control = DeviceView("DVIEWID", share_column=False)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['shareColumn'], False, "CFG share_column should be False")

    def test_device_view_cfg_num_column(self):
        test_control = DeviceView("DVIEWID", num_columns=3)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['numColumns'], 3, "CFG num_column should be 3")

    def test_device_view_cfg_control_title_box_color(self):
        test_control = DeviceView("DVIEWID", control_title_box_color=Color.CHOCOLATE)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(Color(int(cfg_dict['ctrlTitleBoxColor'])), Color.CHOCOLATE, "CFG ctrlTitleBoxColor should be CHOCOLATE")

    def test_device_view_cfg_control_title_box_transparency(self):
        test_control = DeviceView("DVIEWID", control_title_box_transparency=2)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['ctrlTitleBoxTransparency'], 2, "CFG ctrlTitleBoxTransparency should be 2")

    def test_device_view_cfg_control_color(self):
        test_control = DeviceView("DVIEWID", control_color=Color.TOMATO)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(Color(int(cfg_dict['ctrlColor'])), Color.TOMATO, "CFG ctrlColor should be TOMATO")

    def test_device_view_cfg_control_border_color(self):
        test_control = DeviceView("DVIEWID", control_border_color=Color.TURQUOISE)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(Color(int(cfg_dict['ctrlBorderColor'])), Color.TURQUOISE, "CFG ctrlBorderColor should be TURQUOISE")

    def test_device_view_control_background_color(self):
        test_control = DeviceView("DVIEWID", control_background_color=Color.BISQUE)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(Color(int(cfg_dict['ctrlBkgndColor'])), Color.BISQUE, "CFG ctrlBkgndColor should be BISQUE")

    def test_device_view_cfg_control_title_font_size(self):
        test_control = DeviceView("DVIEWID", control_title_font_size=20)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['ctrlTitleFontSize'], 20, "CFG ctrlTitleFontSize should be 20")

    def test_device_view_cfg_control_max_font_size(self):
        test_control = DeviceView("DVIEWID", control_max_font_size=20)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['ctrlMaxFontSize'], 20, "CFG ctrlMaxFontSize should be 20")

    def test_device_view_cfg_control_background_transparency(self):
        test_control = DeviceView("DVIEWID", control_background_transparency=50)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['ctrlBkgndTransparency'], 50, "CFG ctrlBkgndTransparency should be 50")


if __name__ == '__main__':
    unittest.main()



"""The DashIO module is for creating devices, controls and connections for the DashIO app.
"""
from .enums import Color, Icon, Precision, Keyboard, TextAlignment, SliderBarStyle, DialNumberPosition, DialStyle,\
    SoundName, ChartLineType, TimeGraphLineType, TimeGraphPositionOfKey, ButtonState, LabelStyle, KnobStyle,\
    TitlePosition, ChartXAxisLabelsStyle, TextFormat, DirectionStyle, ColorPickerStyle, ControlName
from .audio_visual_display import AudioVisualDisplay
from .chart import Chart, ChartLine, ChartConfig
from .slider import Slider, SliderConfig
from .textbox import TextBox, TextBoxConfig
from .button import Button, ButtonConfig
from .time_graph import TimeGraph, TimeGraphLine, DataPoint, TimeGraphConfig
from .knob import Knob, KnobConfig
from .dial import Dial, DialConfig
from .direction import Direction, DirectionConfig
from .map import Map, MapLocation
from .alarm import Alarm
from .menu import Menu, MenuConfig
from .selector import Selector
from .label import Label, LabelConfig
from .device_view import DeviceView, DeviceViewConfig
from .control import Control, ControlPosition, ControlConfig
from .button_group import ButtonGroup, ButtonGroupConfig
from .event_log import EventLog, EventData
from .color_picker import ColorPicker, ColorPickerConfig



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import datetime
import json
import dateutil.parser
from ..constants import BAD_CHARS
from .control import Control, ControlPosition, ControlConfig, _get_title_position
from .enums import Color, TitlePosition
from .ring_buffer import RingBuffer
from .event import Event

class EventData:
    """Event log data point
    """
    def __init__(self, lines: str, color=Color.WHITE):
        """EventLog data point

        Parameters
        ----------

        lines : str max 25 lines long. Each line is seperated by '\n'
        color : Color, optional
            The color to display this data point on the iotdashboard app, by default Color.WHITE
        """
        self.color = color
        self.timestamp = datetime.datetime.utcnow().replace(microsecond=0, tzinfo=datetime.timezone.utc)
        self.event_lines = []
        no_l = 0

        for s_line in lines.split("\n"):
            self.event_lines.append(s_line.translate(BAD_CHARS))
            no_l += 1
            if no_l > 25:
                break

    def to_json(self):
        """Convert to JSON

        Returns
        -------
        str
            json representation
        """
        event_dict = {
            'time': self.timestamp.isoformat(),
            'color': str(self.color.value),
            'lines': self.event_lines
        }
        return json.dumps(event_dict)

    def __str__(self):
        header = f"{self.timestamp.isoformat()}\t{str(self.color.value)}"
        for line in self.event_lines:
            header += f"\t{line}"
        header += "\n"
        return header


class EventLog(Control):
    """EventLog control
    """

    def __init__(self,
                 control_id: str,
                 title="An Event Log",
                 title_position=TitlePosition.BOTTOM,
                 control_position=None,
                 max_log_entries=100):
        """An EventLog control

        Parameters
        ----------

        control_id : str
            A unique identifier for this control
        title : str, optional
            The title for this control will be displayed on the iotdashboard app, by default "An Event Log"
        title_position : TitlePosition, optional
            The position of the title, by default TitlePosition.BOTTOM
        control_position : ControlPosition, optional
            The position of the control on a DeviceView, by default None
        max_log_entries : int, optional
            The EventLog usues a ring buffer for data entries this dfines the number of
            entries before over wrting older entires, by default 100
        """
        super().__init__("LOG", control_id)
        self._cfg_columnar.append(ControlConfig(control_id, title, control_position, title_position))

        self._message_rx_event = Event()
        self._message_rx_event += self._get_log_from_timestamp
        self.log = RingBuffer(max_log_entries)

    @classmethod
    def from_cfg_dict(cls, cfg_dict: dict):
        """Instatiates Menu from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        Menu
        """
        tmp_cls = cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            _get_title_position(cfg_dict["titlePosition"]),
            ControlPosition(cfg_dict["xPositionRatio"], cfg_dict["yPositionRatio"], cfg_dict["widthRatio"], cfg_dict["heightRatio"])
        )
        tmp_cls.parent_id = cfg_dict["parentID"]
        return tmp_cls

    def _get_log_from_timestamp(self, msg):
        data_str = ""
        try:
            dashboard_id = msg[3]
            from_date = dateutil.parser.isoparse(msg[4])
        except (IndexError, ValueError):
            return ""
        data_str += self._control_hdr_str + dashboard_id
        has_data = False
        for log in self.log.get():
            if log.timestamp > from_date:
                data_str += "\t" + log.to_json()
                has_data = True
        if has_data:
            return data_str

    def add_event_data(self, data: EventData):
        """Add a data point to the log and send it to any connected iotdashboard app

        Parameters
        ----------
        data : EventData
            The data to add to the log
        """
        if isinstance(data, EventData):
            self.log.append(data)
            self.state_str = self._control_hdr_str + str(data)

    def send_data(self):
        """Send the latest log entry to any connected iotdashboard app.
        """
        if self.log:
            self.state_str = self._control_hdr_str + str(self.log.get_latest())



import json
import unittest

from dashio import Knob
from dashio.iotcontrol.enums import Color


class TestKnob(unittest.TestCase):
    def _get_cfg_dict(self, cfg_list: list):
        json_str = cfg_list[0].rpartition('\t')[2]
        return json.loads(json_str)

    def test_knob_control_type(self):
        test_control = Knob("KNOBID")
        test_str_list = test_control._control_hdr_str.split('\t')
        self.assertEqual(test_str_list[2], 'KNOB', "control type should be KNOB")

    def test_knob_control_id(self):
        test_control = Knob("KNOBID")
        test_str_list = test_control._control_hdr_str.split('\t')
        self.assertEqual(test_str_list[3], 'KNOBID', "control type should be KNOBID")

    def test_knob_cfg_min(self):
        test_control = Knob("KNOBID", dial_min=10)
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['min'], 10, "CFG min should be 10")

    def test_knob_cfg_max(self):
        test_control = Knob("KNOBID", dial_max=10)
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['max'], 10, "CFG max should be 10")

    def test_knob_cfg_red_value(self):
        test_control = Knob("KNOBID", red_value=10)
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['redValue'], 10, "CFG redValue should be 10")

    def test_knob_cfg_show_min_max(self):
        test_control = Knob("KNOBID", show_min_max=True)
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['showMinMax'], True, "CFG showMinMax should be True")

    def test_knob_cfg_send_only_on_release(self):
        test_control = Knob("KNOBID", send_only_on_release=True)
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['sendOnlyOnRelease'], True, "CFG sendOnlyOnRelease should be True")

    def test_knob_cfg_dial_follows_knob(self):
        test_control = Knob("KNOBID", dial_follows_knob=True)
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['dialFollowsKnob'], True, "CFG dialFollowsKnob should be True")

    def test_knob_cfg_dial_color(self):
        test_control = Knob("KNOBID", dial_color=Color.AQUA)
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(Color(int(cfg_dict['dialColor'])), Color.AQUA, "CFG dialColor should be AQUA")

    def test_knob_cfg_knob_color(self):
        test_control = Knob("KNOBID", knob_color=Color.AQUAMARINE)
        cfg_dict = self._get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(Color(int(cfg_dict['knobColor'])), Color.AQUAMARINE, "CFG knobColor should be AQUA")


if __name__ == '__main__':
    unittest.main()



import json
import unittest

from dashio import Slider
from dashio.iotcontrol.enums import Color, SliderBarStyle


def _get_cfg_dict(cfg_list: list):
    json_str = cfg_list[0].rpartition('\t')[2]
    return json.loads(json_str)


class TestSlider(unittest.TestCase):

    def test_slider_single_bar_control_type(self):
        test_control = Slider("SLIDERID")
        test_str_list = test_control._control_hdr_str.split('\t')
        self.assertEqual(test_str_list[2], 'SLDR', "control type should be SLCTR")

    def test_slider_single_bar_control_id(self):
        test_control = Slider("SLIDERID")
        test_str_list = test_control._control_hdr_str.split('\t')
        self.assertEqual(test_str_list[3], 'SLIDERID', "control type should be SLIDERID")

    def test_slider_single_bar_cfg_min(self):
        test_control = Slider("SLIDERID", bar_min=10)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['min'], 10, "CFG min should be 10")

    def test_slider_single_bar_cfg_max(self):
        test_control = Slider("SLIDERID", bar_max=10)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['max'], 10, "CFG max should be 10")

    def test_slider_single_bar_cfg_red_value(self):
        test_control = Slider("SLIDERID", red_value=10)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['redValue'], 10, "CFG redValue should be 10")

    def test_slider_single_bar_cfg_show_min_max(self):
        test_control = Slider("SLIDERID", show_min_max=True)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['showMinMax'], True, "CFG showMinMax should be 10")

    def test_slider_single_bar_cfg_slider_enabled(self):
        test_control = Slider("SLIDERID", slider_enabled=True)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['sliderEnabled'], True, "CFG sliderEnabled should be True")

    def test_slider_single_bar_cfg_send_only_on_release(self):
        test_control = Slider("SLIDERID", send_only_on_release=True)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['sendOnlyOnRelease'], True, "CFG sendOnlyOnRelease should be True")

    def test_slider_single_bar_cfg_bar_follows_slider(self):
        test_control = Slider("SLIDERID", bar_follows_slider=True)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(cfg_dict['barFollowsSlider'], True, "CFG barFollowsSlider should be True")

    def test_slider_single_bar_cfg_bar_color(self):
        test_control = Slider("SLIDERID", bar_color=Color.CHARTREUSE)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(Color(int(cfg_dict['barColor'])), Color.CHARTREUSE, "CFG barColor should be True")

    def test_slider_single_bar_cfg_bar_style(self):
        test_control = Slider("SLIDERID", bar_style=SliderBarStyle.SEG)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(SliderBarStyle(cfg_dict['barStyle']), SliderBarStyle.SEG, "CFG barStyle should be SEGMENTED")

    def test_slider_single_bar_cfg_knob_color(self):
        test_control = Slider("SLIDERID", knob_color=Color.CORAL)
        cfg_dict = _get_cfg_dict(test_control.get_cfg(["DEVICEID", "CONTROLID", "DASHID", 1]))
        self.assertEqual(Color(int(cfg_dict['knobColor'])), Color.CORAL, "CFG barStyle should be CORAL")


if __name__ == '__main__':
    unittest.main()



"""action_station.py

Copyright (c) 2022, DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import json
import logging
import threading
import time
import sys
import shortuuid
import zmq


from .constants import CONNECTION_PUB_URL, MEMORY_REQ_URL, TASK_PULL
from .action_station_services.task_service import TaskService
from .action_station_services.timer_service import TimerService, make_timer_config
from .action_station_services.as_servicel import make_test_config
from .action_station_services.modbus_service import ModbusService, make_modbus_config
from .action_station_services.clock_servicel import ClockService, make_clock_config
from .load_config import CONTROL_INSTANCE_DICT, CONFIG_INSTANCE_DICT, decode_cfg64, encode_cfg64


class ActionStation(threading.Thread):
    """_summary_

    Parameters
    ----------
    threading : _type_
        _description_
    """
    def _service_actn_commands(self, payload: bytearray, msg_from: bytearray):
        data = str(payload, "utf-8").strip()
        data_array = data.split("\t")
        try:
            command = data_array[2]
        except (KeyError, IndexError):
            return
        if command in self._action_station_commands:
            reply = self._action_station_commands[command](data_array)
            self.tx_zmq_pub.send_multipart([msg_from, reply.encode()])

    def save_action(self, filename: str, actions_dict: dict):
        """_summary_

        Parameters
        ----------
        filename : str
            Save the Actions
        """
        with open(filename, 'w', encoding='ASCII') as outfile:
            json.dump(actions_dict, outfile, indent=4)

    def load_action(self, filename: str):
        """load action_station from a file.

        Parameters
        ----------
        filename : str
            Filename of the file where the Actions are stored.
        """
        action_station_dict = {}
        try:
            with open(filename, 'r', encoding='ASCII') as infile:
                action_station_dict = json.load(infile)

        except FileNotFoundError:
            pass
        return action_station_dict

    def register_connection(self, connection):
        """Cennections register here"""
        if connection.zmq_connection_uuid not in self.connections_list:
            logging.debug("AS REG CONNECTION")
            self.connections_list.append(connection.zmq_connection_uuid)
            self.connection_zmq_sub.connect(CONNECTION_PUB_URL.format(id=connection.zmq_connection_uuid))
            connection.rx_zmq_sub.connect(CONNECTION_PUB_URL.format(id=self.zmq_connection_uuid))
            self._connect_all_ext_devices()

    def _rm_old_action_station_gui_controls(self):
        old_controls = self._action_station_layout
        if old_controls is not None:
            old_cfg_dict = decode_cfg64(old_controls["provisioning"])
            for control_type, control_list in old_cfg_dict.items():
                if isinstance(control_list, list):
                    for control in control_list:
                        logging.debug("Deleting: %s, %s", control_type, control['controlID'])
                        key = f"{control_type}\t{control['controlID']}"
                        try:
                            del self.device.controls_dict[key]
                            if control_type == 'DVVW':
                                self.device.number_of_device_views = self.device.number_of_device_views - 1
                        except KeyError:
                            logging.debug("Error deleting control: %s, %s", control_type, control['controlID'])

    def _delete_gui_configs(self, cfg_dict: dict):
        modified = False
        for control_type, control_list in cfg_dict.items():
            if isinstance(control_list, list):
                for control in control_list:
                    key = f"{control_type}\t{control['controlID']}"
                    if self.device.is_control_loaded(control_type, control['controlID']):
                        modified = True
                        logging.debug("Deleting layouts: %s, %s", control_type, control['controlID'])
                        try:
                            self.device.controls_dict[key].del_configs_columnar()
                        except KeyError:
                            logging.debug("Error deleting layout: %s, %s", control_type, control['controlID'])
        return modified

    def _update_gui_controls(self, cfg_dict: dict):
        new_cfg_dict = {}
        new_cfg_dict["CFG"] = self.device._cfg
        added_control_ids = []
        modified = False
        for control_type, control_list in cfg_dict.items():
            if isinstance(control_list, list):
                for control in control_list:
                    key = f"{control_type}\t{control['controlID']}"
                    if self.device.is_control_loaded(control_type, control['controlID']):
                        # load new config into control
                        logging.debug("Adding layouts: %s, %s", control_type, control['controlID'])
                        cfg = CONFIG_INSTANCE_DICT[control_type].from_dict(control)
                        self.device.controls_dict[key].add_config_columnar(cfg)
                        modified = True
                        if key in added_control_ids:
                            # Add the next config to the control
                            new_cfg_dict[control_type].append(control)
                    else:
                        # Create a new control.
                        g_control = CONTROL_INSTANCE_DICT[control_type].from_cfg_dict(control)
                        self.device.add_control(g_control)
                        added_control_ids.append(key)
                        if control_type not in new_cfg_dict:
                            new_cfg_dict[control_type] = []
                        logging.debug("Added control: %s", control_type + ":" + control["controlID"])
                        new_cfg_dict[control_type].append(control)
                        modified = True
        return modified, new_cfg_dict

    def load_all_controls_from_config(self, cfg_dict) -> dict:
        """Loads all the controls in cfg_dict into device and returns a dictionary of the control objects

        Parameters
        ----------
        device : Dashio.Device
            The device to attach the controls to
        cfg_dict : Dict
            dictionary of the CFG loaded by decode_cfg from a CFG64 or json

        Returns
        -------
        dict
            Dictionary of the control objects
        """

        new_c64 = ""
        modified = False
        modified = self._delete_gui_configs(cfg_dict)
        self._rm_old_action_station_gui_controls()
        modified, new_cfg_dict = self._update_gui_controls(cfg_dict)
        if modified:
            self.device.inc_config_revision()
        if new_cfg_dict:
            # logging.debug("New C64:\n%s", json.dumps(new_cfg_dict, indent=4))
            new_c64 = encode_cfg64(new_cfg_dict)
        return new_c64

    def add_gui_controls(self, g_object: dict):
        """Add a GUI control"""
        cfg_dict = decode_cfg64(g_object["provisioning"])
        new_c64 = self.load_all_controls_from_config(cfg_dict)
        g_object["provisioning"] = new_c64
        self._action_station_layout = g_object
        # there can only be one C64

    def close(self):
        """Close the action_station json filename
        """
        self.save_action(self._json_filename, self.action_station_dict)
        for control in self.thread_dicts.values():
            control.close()
        self.running = False

    def _connect_all_ext_devices(self):
        for device_id in self.remote_device_ids:
            self._connect_device_id(device_id)

    def _connect_device_id(self, device_id):
        msg_dict = {
            'msgType': 'connect',
            'deviceID': device_id
        }
        logging.debug("AS CONNECT: %s", device_id)
        self.tx_zmq_pub.send_multipart([b"COMMAND", json.dumps(msg_dict).encode()])

    def _disconnect_device_id(self, device_id):
        msg_dict = {
            'msgType': 'disconnect',
            'deviceID': device_id
        }
        logging.debug("AS DISCONNECT: %s", device_id)
        self.tx_zmq_pub.send_multipart([b"COMMAND", json.dumps(msg_dict).encode()])

    def _start_control(self, t_object: dict):
        if t_object['uuid'] in self.thread_dicts:
            self.thread_dicts[t_object['uuid']].close()
            time.sleep(0.1)
        logging.debug("INIT TASK %s", t_object['uuid'])
        self.thread_dicts[t_object['uuid']] = self.service_objects_defs[t_object['objectType']](self.device_id, self.zmq_service_uuid, t_object, self.context)
        return True

    def _list_command(self, data):
        j_object_list = []
        for j_object in self.configured_services.values():
            action_pair = {
                "name": j_object['name'],
                "uuid": j_object['uuid'],
                "objectType": j_object['objectType'],
                "revision": j_object.get('revision', 1)
            }
            j_object_list.append(action_pair)
        # TODO delete this for loop
        for j_object in self.configs.values():
            action_pair = {
                "name": j_object['name'],
                "uuid": j_object['uuid'],
                "objectType": j_object['objectType'],
                "revision": j_object.get('revision', 1)
            }
            j_object_list.append(action_pair)
        result = {
            'objectType': "LIST_RESULT",
            'list': j_object_list
        }
        reply = f"\t{self.device_id}\tACTN\tLIST\t{json.dumps(result)}\n"
        return reply

    def _list_configs_command(self, data):
        j_object_list = []
        for j_object in self.configs.values():
            if j_object['objectType'] == "CONFIG":
                j_object_list.append(j_object)
        result = {
            'objectType': "LIST_RESULT",
            'list': j_object_list
        }
        reply = f"\t{self.device_id}\tACTN\tLIST_CONFIGS\t{json.dumps(result)}\n"
        return reply

    def _list_tasks_command(self, data):
        j_object_list = []
        for j_object in self.configured_services.values():
            if j_object['objectType'] == "TASK":
                action_pair = {
                    "name": j_object['name'],
                    "uuid": j_object['uuid'],
                    "objectType": j_object['objectType'],
                    "revision": j_object.get('revision', 1)
                }
                j_object_list.append(action_pair)
        result = {
            'objectType': "LIST_RESULT",
            'list': j_object_list
        }
        reply = f"\t{self.device_id}\tACTN\tLIST_TASKS\t{json.dumps(result)}\n"
        return reply

    def _get_command(self, data):
        payload = json.loads(data[3])
        j_object = self.configured_services.get(payload["uuid"], {})
        reply = f"\t{self.device_id}\tACTN\tGET\t{json.dumps(j_object)}\n"
        return reply

    def _delete_command(self, data):
        logging.debug("DELETE: %s", data)
        payload = json.loads(data[3])
        result = {
            'objectType': "DELETE_RESULT",
            'uuid': payload['uuid'],
            'result': False
        }
        try:
            store_obj = self.configured_services[payload["uuid"]]
            if store_obj['objectType'] in self.service_objects_defs:
                self.thread_dicts[payload['uuid']].close()
                time.sleep(0.1)
                del self.thread_dicts[payload['uuid']]
            if store_obj['objectType'] == "DVCE_CONFIG":
                self._rm_old_action_station_gui_controls()
                self._action_station_layout = None
            del self.configured_services[payload["uuid"]]

            result['result'] = True
        except KeyError as error:
            logging.debug("Key Error: %s", error)
            result['result'] = False
        reply = f"\t{self.device_id}\tACTN\tDELETE\t{json.dumps(result)}\n"
        self.save_action(self._json_filename, self.action_station_dict)
        return reply

    def _rx_command(self, msg):
        msg_dict = json.loads(msg)
        logging.debug("AS RX CMD: %s", msg_dict)

    def _update_command(self, data):
        payload = json.loads(data[3])
        result = {
            'objectType': "UPDATE_RESULT",
            'uuid': payload['uuid'],
            'result': False
        }
        try:
            if payload['objectType'] in self.service_objects_defs:
                self.configured_services[payload['uuid']] = payload
                result['result'] = self._start_control(payload)
            if payload['objectType'] == 'DVCE_CONFIG':
                result['result'] = True
                self.add_gui_controls(payload)
                self.configured_services[payload['uuid']] = payload
        except KeyError as error:
            msg = "UPDATE: payload has no objectType"
            logging.debug("%s, %s", msg, error)
            result['message'] = msg
        if result['result']:
            self.save_action(self._json_filename, self.action_station_dict)
        reply = f"\t{self.device_id}\tACTN\tUPDATE\t{json.dumps(result)}\n"
        return reply

    def _run_command(self, data):
        # payload = json.loads(data[3])
        reply = ""
        return reply

    def __init__(self, device, max_actions=100, number_timers=10, memory_storage_size=0, context: zmq.Context = None):
        """Action Station"""
        threading.Thread.__init__(self, daemon=True)
        self.context = context or zmq.Context.instance()
        self.device_id = device.device_id
        self.device = device
        self._json_filename = f"{self.device_id}_Actions.json"

        self.configured_services = {}
        self.configs = {}
        self.memory_tasks = {}
        self.remote_device_ids = []
        self.connections_list = []
        self._action_station_layout = None
        self.action_station_dict = self.load_action(self._json_filename)
        self.max_actions = max_actions

        self.service_objects_defs = {
            'TASK': TaskService,
            'TMR': TimerService,
            'MDBS': ModbusService,
            'CLK': ClockService
        }
        self.thread_dicts = {}  # For the Instantiated control and task objects.

        self._action_station_commands = {
            "LIST": self._list_command,
            "LIST_TASKS": self._list_tasks_command,
            "LIST_CONFIGS": self._list_configs_command,
            "GET": self._get_command,
            "DELETE": self._delete_command,
            "UPDATE": self._update_command,
            "RUN": self._run_command
        }
        timer_cfg = make_timer_config(number_timers)
        test_cfg = make_test_config(1)
        modbus_cfg = make_modbus_config(1)
        clock_cfg = make_clock_config(1)
        self.configs[timer_cfg['uuid']] = timer_cfg
        self.configs[test_cfg['uuid']] = test_cfg
        self.configs[modbus_cfg['uuid']] = modbus_cfg
        self.configs[clock_cfg['uuid']] = clock_cfg
        self.zmq_service_uuid = "SRVC:" + shortuuid.uuid()
        if not self.action_station_dict:
            self.zmq_connection_uuid = "ACTN:" + shortuuid.uuid()
            self.action_station_dict['actionStationID'] = self.zmq_connection_uuid
            self.action_station_dict['services'] = self.configured_services
            self.action_station_dict['tasksMemory'] = self.memory_tasks
        else:
            try:
                self.zmq_connection_uuid = self.action_station_dict['actionStationID']
                self.configured_services = self.action_station_dict['services']
                self.memory_tasks = self.action_station_dict['tasksMemory']
            except KeyError:
                sys.exit(f"Old json formatted file. Please delete '{self._json_filename}' and restart")

        self.running = True
        self.start()

    def run(self):

        self.zmq_service_pub = self.context.socket(zmq.PUB)

        self.zmq_service_pub.bind(CONNECTION_PUB_URL.format(id=self.zmq_service_uuid))

        self.tx_zmq_pub = self.context.socket(zmq.PUB)
        self.tx_zmq_pub.bind(CONNECTION_PUB_URL.format(id=self.zmq_connection_uuid))

        self.device_zmq_sub = self.context.socket(zmq.SUB)
        #  Subscribe on ALL, and my connection
        self.device_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "ALL")
        self.device_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, self.zmq_connection_uuid)

        #  rx_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "ANNOUNCE")
        self.connection_zmq_sub = self.context.socket(zmq.SUB)
        self.connection_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "COMMAND")
        self.connection_zmq_sub.setsockopt_string(zmq.SUBSCRIBE, "\t")  # - All valid DashIO messaging

        #  Socket to receive messages on
        task_receiver = self.context.socket(zmq.PULL)
        task_receiver.bind(TASK_PULL.format(id=self.zmq_service_uuid))

        memory_socket = self.context.socket(zmq.REP)
        memory_socket.bind(MEMORY_REQ_URL.format(id=self.device_id))

        poller = zmq.Poller()
        poller.register(self.device_zmq_sub, zmq.POLLIN)
        poller.register(self.connection_zmq_sub, zmq.POLLIN)
        poller.register(task_receiver, zmq.POLLIN)
        poller.register(memory_socket, zmq.POLLIN)

        for j_object in self.configured_services.values():
            if j_object['objectType'] == "DVCE_CONFIG":
                self.add_gui_controls(j_object)
            else:
                self._start_control(j_object)

        while self.running:
            try:
                socks = dict(poller.poll(20))
            except zmq.error.ContextTerminated:
                break
            if self.device_zmq_sub in socks:
                try:
                    [_, msg] = self.device_zmq_sub.recv_multipart()
                except ValueError:
                    # If there aren't two parts continue.
                    pass
                if msg:
                    logging.debug("ActionStation Device RX:\n%s", msg.decode().rstrip())
                    self.zmq_service_pub.send_multipart([msg, self.device_id.encode()])

            if self.connection_zmq_sub in socks:
                msg, msg_from = self.connection_zmq_sub.recv_multipart()
                # logging.debug("ActionStation Connection RX:\n%s, %s", msg_from, msg.decode())
                if msg[0] == b"COMMAND":
                    self._rx_command(msg_from)
                    continue
                if msg.startswith(f"\t{self.device_id}\tACTN".encode()):
                    self._service_actn_commands(msg, msg_from)
                    continue
                self.zmq_service_pub.send_multipart([msg, msg_from])

            if task_receiver in socks:
                msg_to, msg = task_receiver.recv_multipart()
                self.tx_zmq_pub.send_multipart([msg_to, msg])
                if msg_to == b'COMMAND':
                    msg_dict = json.loads(msg)
                    if msg_dict['msgType'] == 'connect':
                        if msg_dict['deviceID'] not in self.remote_device_ids:
                            logging.debug("Added remote deviceID: %s", msg_dict['deviceID'])
                            self.remote_device_ids.append(msg_dict['deviceID'])

            if memory_socket in socks:
                message = memory_socket.recv_multipart()
                logging.debug("MEM Rx: %s", message)
                if len(message) == 3:
                    if message[0] == b'SET':
                        self.memory_tasks[message[1].decode()] = message[2].decode()
                        logging.debug("MEM Tx: SET: %s, TO: %s", message[1], message[2])
                        memory_socket.send_multipart([message[0], message[1], message[2]])
                    if message[0] == b'GET':
                        logging.debug("MEM Tx: GET: %s, RTN: %s", message[1], message[1])
                        memory_socket.send_multipart([message[0], message[1], self.memory_tasks[message[1].decode()].encode()])
                #  Send error reply back to client
                else:
                    memory_socket.send_multipart([b'ERROR', b'ERROR', b'ERROR'])

        self.tx_zmq_pub.close()
        self.device_zmq_sub.close()
        memory_socket.close()
        task_receiver.close()



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import datetime
import dateutil
import json

from ..constants import BAD_CHARS
from .control import Control, ControlPosition, ControlConfig, _get_title_position
from .enums import TitlePosition, Color
from .event import Event

class MapLocation:
    """
    A json version of a map location.
    """
    def __init__(self, latitude: str, longitude: str, average_speed=None, peak_speed=None, course=None, altitude=None, distance=None):
        """MapLocation

        Parameters
        ----------
        latitude : str
            The latitude
        longitude : str
            The longitude
        average_speed : str, optional
            Average speed, by default None
        peak_speed : str, optional
            Peak speed, by default None
        course : str, optional
            couse, by default None
        altitude : str, optional
            altitude, by default None
        distance : str, optional
            distance, by default None
        """
        self.timestamp = datetime.datetime.utcnow().replace(microsecond=0, tzinfo=datetime.timezone.utc)
        self._map_loc = {}
        self._map_loc["time"] = self.timestamp.isoformat()
        self._map_loc["latitude"] = latitude.translate(BAD_CHARS)
        self._map_loc["longitude"] = longitude.translate(BAD_CHARS)
        if average_speed:
            self._map_loc["avgeSpeed"] = average_speed
        if peak_speed:
            self._map_loc["peakSpeed"] = peak_speed
        if course:
            self._map_loc["course"] = course
        if altitude:
            self._map_loc["altitude"] = altitude
        if distance:
            self._map_loc["distance"] = distance

    def to_json(self):
        """Make a json representation of a map point.

        Returns
        -------
        str
        """
        return json.dumps(self._map_loc)

    def get_simple_format(self):
        """Returns the simple format of the MAP Location

        Returns
        -------
        _type_
            _description_
        """
        return f"\t{self.latitude},{self.longitude}\n"


class MapTrack:
    """A Map track
    """
    def __init__(self, track_id: str, text="", color=Color.RED) -> None:
        self.track_id = track_id.translate(BAD_CHARS)
        self.text = text.translate(BAD_CHARS)
        self.color = color
        self.track_start_time = datetime.datetime.utcnow().replace(microsecond=0, tzinfo=datetime.timezone.utc)
        self.locations = []

    def add_location(self, location: MapLocation) -> None:
        """Add a location to the Track

        Parameters
        ----------
        location : MapLocation
            The Map location to add to the track.
        """
        if isinstance(location, MapLocation):
            self.locations.append(location)

    def get_track(self) -> str:
        """Returns the track in DashIO long format

        Returns
        -------
        str
            track in DashIO long format
        """
        reply = f"\t{self.track_id}\t{self.text}\t{self.color.value}"
        for loc in self.locations:
            reply += f"\t{loc.map_to_json()}"
        return reply

    def get_last_location(self) -> str:
        """Returns the last track location in short format

        Returns
        -------
        str
            The last location in track in short format
        """
        reply = f"{self.track_id}" + self.locations[-1].get_simple_format()
        return reply


class Map(Control):
    """A Map control
    """

    def __init__(
        self,
        control_id,
        title="A Map",
        title_position=TitlePosition.BOTTOM,
        control_position=None
    ):
        """A Map control

        Parameters
        ----------
        control_id : str
            An unique control identity string. The control identity string must be a unique string for each control per device.
        title : str, optional
            Title of the control, by default "A Map"
        control_position : ControlPosition, optional
            The position of the control on a DeviceView, by default None
        title_position : TitlePosition, optional
            Position of the title when displayed on the iotdashboard app, by default None
        """
        super().__init__("MAP", control_id)
        self._cfg_columnar.append(ControlConfig(control_id, title, control_position, title_position))
        self.tracks = {}
        self.tracks["DEFAULT"] = self.default_track

        self._message_rx_event = Event()
        self._message_rx_event += self._get_tracks_from_timestamp

    @classmethod
    def from_cfg_dict(cls, cfg_dict: dict):
        """Instatiates Map from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        Map
        """
        tmp_cls = cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            _get_title_position(cfg_dict["titlePosition"]),
            ControlPosition(cfg_dict["xPositionRatio"], cfg_dict["yPositionRatio"], cfg_dict["widthRatio"], cfg_dict["heightRatio"])
        )
        tmp_cls.parent_id = cfg_dict["parentID"]
        return tmp_cls

    def _get_tracks_from_timestamp(self, msg):
        reply = ""
        try:
            dashboard_id = msg[3]
            from_timestamp = dateutil.parser.isoparse(msg[4])
        except IndexError:
            return ""
        if self.tracks:
            for _, value in self.tracks:
                if value.track_start_time > from_timestamp:
                    reply = + self._control_hdr_str + dashboard_id + value.get_track() + "\n"
        return reply

    def add_location_to_track(self, location: MapLocation, track_id: str) -> None:
        """Add Location to the map

        Parameters
        ----------
        location : MapLocation.
            The location to add to the map
        track_id : Track ID for the location.
        """
        if track_id in self.tracks:
            self.tracks[track_id].add_location(location)
        else:
            self.tracks[track_id] = MapTrack("track_id")
            self.tracks[track_id].add_location(location)
        self.send_location(location, track_id)

    def send_location(self, location, track_id=""):
        """Sends the locations to the map
        """
        state_str = ""
        state_str += self._control_hdr_str + track_id + location.simple_format()
        self.state_str = state_str



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


class Event:
    """
    Event class that call multiple functions/methods in a set
    """
    def __init__(self):
        self.handlers = set()

    def handle(self, handler):
        """Add handler to set

        Parameters
        ----------
        handler : function
            A function or method to be added to the set

        Returns
        -------
        Event
            Itself with the added handler
        """
        self.handlers.add(handler)
        return self

    def unhandle(self, handler):
        """Removes the handler from the set

        Parameters
        ----------
        handler : function or method
            A function or method to be removed from the set

        Returns
        -------
        Event
            Itself with the handler removed

        Raises
        ------
        ValueError
            IF the handler isn't in the set
        """
        try:
            self.handlers.remove(handler)
        except KeyError as not_in_set:
            raise ValueError("Handler is not handling this event, so cannot unhandle it.") from not_in_set
        return self

    def fire(self, *args, **kargs):
        """Runs all the handlers
        """
        for handler in self.handlers:
            handler(*args, **kargs)

    def get_handler_count(self):
        """Returns the number of handlers

        Returns
        -------
        int
            The number of handlers
        """
        return len(self.handlers)

    __iadd__ = handle
    __isub__ = unhandle
    __call__ = fire
    __len__ = get_handler_count



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from .control import Control, ControlConfig, _get_color, _get_icon, _get_device_view_style
from .enums import Color, Icon, DeviceViewStyle


class DeviceViewConfig(ControlConfig):
    """DeviceViewConfig"""
    def __init__(
        self,
        control_id: str,
        title: str,
        style: DeviceViewStyle,
        icon_name: Icon,
        color: Color,
        share_column: bool,
        num_columns: int,
        control_title_box_color: Color,
        control_title_box_transparency: int,
        control_color: Color,
        control_border_color: Color,
        control_background_color: Color,
        control_title_font_size: int,
        control_max_font_size: int,
        control_background_transparency: int,
        num_grid_columns: int,
        num_grid_rows: int,
    ) -> None:
        super().__init__(control_id, title, control_position=None, title_position=None)
        self.cfg["iconName"] = icon_name.value
        self.cfg["style"] = str(style.value)
        self.cfg["color"] = str(color.value)
        self.cfg["shareColumn"] = share_column
        if 1 <= num_columns <= 10:
            self.cfg["numColumns"] = num_columns
        else:
            raise ValueError("num_columns must be in the range 1 to 10")
        self.cfg["ctrlTitleBoxColor"] = str(control_title_box_color.value)
        if 0 <= control_title_box_transparency <= 100:
            self.cfg["ctrlTitleBoxTransparency"] = control_title_box_transparency
        else:
            raise ValueError("Value must be in the range 0 to 100")
        self.cfg["ctrlColor"] = str(control_color.value)
        self.cfg["ctrlBorderColor"] = str(control_border_color.value)
        self.cfg["ctrlBkgndColor"] = str(control_background_color.value)
        self.cfg["ctrlTitleFontSize"] = control_title_font_size
        self.cfg["ctrlMaxFontSize"] = control_max_font_size
        if 0 <= control_background_transparency <= 100:
            self.cfg["ctrlBkgndTransparency"] = control_background_transparency
        else:
            raise ValueError("Value must be in the range 0 to 100")
        self.cfg["gridColumns"] = num_grid_columns
        self.cfg["gridRows"] = num_grid_rows

    @classmethod
    def from_dict(cls, cfg_dict: dict):
        """Instantiates DeviceViewConfig from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        DeviceViewConfig
        """
        tmp_cls = cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            _get_device_view_style(cfg_dict["style"]),
            _get_icon(cfg_dict["iconName"]),
            _get_color(cfg_dict["color"]),
            cfg_dict["shareColumn"],
            cfg_dict["numColumns"],
            _get_color(cfg_dict["ctrlTitleBoxColor"]),
            cfg_dict["ctrlTitleBoxTransparency"],
            _get_color(cfg_dict["ctrlColor"]),
            _get_color(cfg_dict["ctrlBorderColor"]),
            _get_color(cfg_dict["ctrlBkgndColor"]),
            cfg_dict["ctrlTitleFontSize"],
            cfg_dict["ctrlMaxFontSize"],
            cfg_dict["ctrlBkgndTransparency"],
            cfg_dict["gridColumns"],
            cfg_dict["gridRows"]
        )
        return tmp_cls


class DeviceView(Control):
    """A DeviceView provides a control that descibes appearance and style of the group of controls
        that are displayed on this DeviceView by the iotdashboard app.

    Attributes
    ----------
    control_id : str
       An unique control identity string. The control identity string must be a unique string for each control per device
       Each control inherits the DeviceViews Control settings
    title : str
        The controls title
    style : DeviceViewStyle
        The style of the DeviceView
    icon : Icon
        The Icon representing the DeviceView
    color : Color
        The color of the DeviceView
    share_column : Booloan
        Something something
    num_columns : int
        Number of columns for DeviceView
    control_title_box_color : Color
        Title box color for controls
    control_title_box_transparency : int
        Title box transparency for controls
    num_grid_columns : int
        The num of grid columns on the edit view
    num_grid_rows : int
        The num of grid rows on the edit view


    Methods
    -------
    add_control(Control) :
        Add a control to the device view

    from_cfg_dict(dict) :
        A class method to instantiate a DeviceView object from a dictionary
    """

    def __init__(
        self,
        control_id,
        title="A DeviceView",
        style=DeviceViewStyle.BASIC,
        icon=Icon.SQUARE,
        color=Color.BLACK,
        share_column=True,
        num_columns=1,
        control_title_box_color=Color.BLACK,
        control_title_box_transparency=0,
        control_color=Color.WHITE_SMOKE,
        control_border_color=Color.WHITE_SMOKE,
        control_background_color=Color.BLACK,
        control_title_font_size=16,
        control_max_font_size=20,
        control_background_transparency=0,
        num_grid_columns=22,
        num_grid_rows=32
    ):
        super().__init__("DVVW", control_id)
        self._cfg_columnar.append(
            DeviceViewConfig(
                control_id,
                title,
                style,
                icon,
                color,
                share_column,
                num_columns,
                control_title_box_color,
                control_title_box_transparency,
                control_color,
                control_border_color,
                control_background_color,
                control_title_font_size,
                control_max_font_size,
                control_background_transparency,
                num_grid_columns,
                num_grid_rows
            )
        )
        self._state_str = ""

    @classmethod
    def from_cfg_dict(cls, cfg_dict: dict):
        """Instatiates DeviceView from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        DeviceView
        """
        return cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            _get_device_view_style(cfg_dict["style"]),
            _get_icon(cfg_dict["iconName"]),
            _get_color(cfg_dict["color"]),
            cfg_dict["shareColumn"],
            cfg_dict["numColumns"],
            _get_color(cfg_dict["ctrlTitleBoxColor"]),
            cfg_dict["ctrlTitleBoxTransparency"],
            _get_color(cfg_dict["ctrlColor"]),
            _get_color(cfg_dict["ctrlBorderColor"]),
            _get_color(cfg_dict["ctrlBkgndColor"]),
            cfg_dict["ctrlTitleFontSize"],
            cfg_dict["ctrlMaxFontSize"],
            cfg_dict["ctrlBkgndTransparency"],
            cfg_dict["gridColumns"],
            cfg_dict["gridRows"]
        )

    def add_control(self, control):
        """Add a control to the DeviceView

        Parameters
        ----------
        control : Control
            The control to add
        """
        control.parent_id = self.control_id



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="dashio",
    version="3.3.2",
    description="DashIO interface library",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="James Boulton",
    author_email="james@dashio.com",
    url="https://dashio.io",
    download_url="https://github.com/dashio-connect/python-dashio",
    packages=find_packages(),
    license="MIT",
    classifiers=["Programming Language :: Python :: 3.6", "Operating System :: OS Independent"],
    install_requires=["paho-mqtt", "pyzmq", "python-dateutil", "zeroconf", "shortuuid", "pyserial", "astral"],
    python_requires='>3.6.0',
)



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import json
import logging
from ..constants import BAD_CHARS
from .enums import ColorPickerStyle, DeviceViewStyle, DialNumberPosition, DirectionStyle, ChartXAxisLabelsStyle, Keyboard, KnobStyle, Precision,\
    TextAlignment, TitlePosition, Icon, Color, TextFormat, LabelStyle, SliderBarStyle, DialStyle, TimeGraphPositionOfKey
from .event import Event


def _get_icon(icon_str: str) -> Icon:
    icon_name = icon_str.upper().replace(" ", "")
    return Icon[icon_name]


def _get_color(color_str: str) -> Color:
    color_name = color_str.upper().replace(" ", "_")
    return Color[color_name]


def _get_title_position(position_str: str) -> TitlePosition:
    t_pos_name = position_str.upper().replace(" ", "")
    return TitlePosition[t_pos_name]


def _get_text_align(text_align_str: str) -> TextAlignment:
    t_align_name = text_align_str.upper().replace(" ", "")
    return TextAlignment[t_align_name]


def _get_text_format(text_format_str: str) -> TextFormat:
    t_format_name = text_format_str.upper().replace(" ", "")
    return TextFormat[t_format_name]


def _get_precision(precision_int: int) -> Precision:
    return Precision(precision_int)


def _get_keyboard_type(keyboard_str: str) -> Keyboard:
    t_keyboard = keyboard_str.upper().replace(" ", "")
    return Keyboard[t_keyboard]


def _get_device_view_style(device_vuew_style: str) -> DeviceViewStyle:
    dvs_name = device_vuew_style.upper().replace(" ", "")
    return DeviceViewStyle[dvs_name]


def _get_color_picker_style(color_picker_style: str) -> ColorPickerStyle:
    cps_name = color_picker_style.upper().replace(" ", "")
    return ColorPickerStyle[cps_name]


def _get_dial_number_position(dn_position_str: str) -> DialNumberPosition:
    dnp_name = dn_position_str.upper().replace(" ", "")
    return DialNumberPosition[dnp_name]


def _get_direction_style(dir_style_str: str) -> DirectionStyle:
    ds_name = dir_style_str.upper().replace(" ", "")
    return DirectionStyle[ds_name]


def _get_chart_x_axis_labels_style(gxal_style_str: str) -> ChartXAxisLabelsStyle:
    ds_name = gxal_style_str.upper().replace(" ", "")
    return ChartXAxisLabelsStyle[ds_name]


def _get_knob_style(knob_style_str: str) -> KnobStyle:
    ks_name = knob_style_str.upper().replace(" ", "")
    return KnobStyle[ks_name]


def _get_label_style(label_style_str: str) -> LabelStyle:
    ks_name = label_style_str.upper().replace(" ", "")
    return LabelStyle[ks_name]


def _get_bar_style(bar_style_str: str) -> SliderBarStyle:
    bs_name = bar_style_str.upper().replace(" ", "")
    return SliderBarStyle[bs_name]


def _get_dial_style(dial_style_str: str) -> DialStyle:
    bs_name = dial_style_str.upper().replace(" ", "")
    return DialStyle[bs_name]


def _get_time_graph_position_of_key(tgp_of_key: str) -> TimeGraphPositionOfKey:
    tgp_of_key = tgp_of_key.upper().replace(" ", "")
    return TimeGraphPositionOfKey[tgp_of_key]


class ControlPosition:
    """
    ControlPosition
        Used to describe a controls position.
    """
    def __init__(self, x_position_ratio: float, y_position_ratio: float, width_ratio: float, height_ratio: float):
        """The ControlPosition class describes the location and size of a control on a DeviceView. The
        x_postion and y_position ratio place the top left hand corner of the control. The width and height ratio
        describe the controls size. The ratio is a number betwwen 0 and 1 representing the width and height of the
        DeviceView.

        Parameters
        ----------
        x_position_ratio : float
            Left side position expressed as a ration between 0 and 1
        y_position_ratio : float
            Upper side position expressed as a ration between 0 and 1
        width_ratio : float
            Control width expressed as a ration between 0 and 1
        height_ratio : float
            Control height expressed as a ration between 0 and 1
        """
        self.x_position_ratio = x_position_ratio
        self.y_position_ratio = y_position_ratio
        self.width_ratio = width_ratio
        self.height_ratio = height_ratio


class ControlConfig:
    """Base COntrolConfig"""
    def get_cfg_json(self) -> str:
        """Returns the CFG str for the control called when the iotdashboard app asks for a CFG

        Parameters
        ----------
        data : list from IoTDashboard
            The command from IoTDashboard split on \t into a list

        Returns
        -------
        str
            The CFG for this control
        """
        cfg_str = json.dumps(self.cfg) + "\n"

        return cfg_str

    def get_cfg64(self) -> dict:
        """Returns the CFG dict for the control called when the iotdashboard app asks for a CFG

        Parameters
        ----------
        data : list from IoTDashboard
            The command from IoTDashboard split on \t into a list

        Returns
        -------
        dict
            The CFG dict for this control
        """
        return self.cfg

    def update_position(self, control_position: ControlPosition):
        """Updates the controls position from control_position

        Parameters
        ----------
        control_position : ControlPosition
            The position to use to update the control
        """
        self.cfg["xPositionRatio"] = control_position.x_position_ratio
        self.cfg["yPositionRatio"] = control_position.y_position_ratio
        self.cfg["widthRatio"] = control_position.width_ratio
        self.cfg["heightRatio"] = control_position.height_ratio

    def __init__(
        self,
        control_id: str,
        title: str,
        control_position: ControlPosition,
        title_position: TitlePosition

    ) -> None:
        self.cfg = {}
        if title is not None:
            self.cfg["title"] = title.translate(BAD_CHARS)
        self._title_position = None
        if title_position is not None:
            self.cfg["titlePosition"] = title_position.value
        if control_position is not None:
            self.cfg["xPositionRatio"] = control_position.x_position_ratio
            self.cfg["yPositionRatio"] = control_position.y_position_ratio
            self.cfg["widthRatio"] = control_position.width_ratio
            self.cfg["heightRatio"] = control_position.height_ratio
        self.cfg["controlID"] = control_id.translate(BAD_CHARS)
        self.cfg["parentID"] = ""

    @classmethod
    def from_dict(cls, cfg_dict: dict):
        """Instatiates Dial from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        ControlConfig
        """

        tmp_cls = cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            ControlPosition(cfg_dict["xPositionRatio"], cfg_dict["yPositionRatio"], cfg_dict["widthRatio"], cfg_dict["heightRatio"]),
            _get_title_position(cfg_dict["titlePosition"])
        )
        tmp_cls.parent_id = cfg_dict["parentID"]
        return tmp_cls

    @property
    def parent_id(self) -> str:
        """The parent control or deviceview this control belongs to

        Returns
        -------
        str
            The parent_id
        """
        return self.cfg["parentID"]

    @parent_id.setter
    def parent_id(self, val: str):
        _val = val.translate(BAD_CHARS)
        self.cfg["parentID"] = _val


class Control():
    """Base class for controls. """

    def get_state(self) -> str:
        """This is called by iotdashboard app. Controls need to implement their own version."""
        return ""

    def get_cfg(self, data) -> list:
        """Returns the CFG str for the control called when the iotdashboard app asks for a CFG

        Parameters
        ----------
        data : list from IoTDashboard
            The command from IoTDashboard split on \t into a list

        Returns
        -------
        str
            The CFG for this control
        """
        cfg_list = []
        try:
            num_columns = int(data[3])
            dashboard_id = data[2]
        except (IndexError, ValueError):
            return cfg_list
        if num_columns >= self._cfg_full_page_no_columns and self._cfg_full_page:
            for cfg in self._cfg_full_page:
                cfg_list.append(f"\tCFG\t{dashboard_id}\t{self.cntrl_type}\t{cfg.get_cfg_json()}")
        else:
            for control_cfg in self._cfg_columnar:
                cfg_list.append(f"\tCFG\t{dashboard_id}\t{self.cntrl_type}\t{control_cfg.get_cfg_json()}")
        return cfg_list

    def get_cfg64(self, data) -> list:
        """Returns the CFG dict for the control called when the iotdashboard app asks for a CFG

        Parameters
        ----------
        data : list from IoTDashboard
            The command from IoTDashboard split on \t into a list

        Returns
        -------
        dict
            The CFG dict for this control
        """
        cfg_list = []
        try:
            num_columns = int(data[3])
        except (IndexError, ValueError):
            return []
        if num_columns >= self._cfg_full_page_no_columns and self._cfg_full_page:
            for cfg in self._cfg_full_page:
                cfg_list.append(cfg.get_cfg64())
        else:
            for cfg in self._cfg_columnar:
                cfg_list.append(cfg.get_cfg64())
        return cfg_list

    def add_config_columnar(self, config):
        """Add a duplicate Config for the columnar view"""
        config.cfg["controlID"] = self.control_id
        self._cfg_columnar.append(config)

    def add_config_full_page(self, config):
        """Add a duplicate Config for the full page view"""
        config.cfg["controlID"] = self.control_id
        self._cfg_full_page.append(config)

    def set_no_culumns_full_page(self, no_columns: int):
        """Set the number of columns that the full page view uses."""
        self._cfg_full_page_no_columns = no_columns

    def add_receive_message_callback(self, callback):
        """Add a callback to receive incoming messages to the control."""
        self._message_rx_event += callback

    def remove_receive_message_callback(self, callback):
        """Remaove a callback from receive incoming messages to the control."""
        self._message_rx_event -= callback

    def add_transmit_message_callback(self, callback):
        """Add a callback for transmitted messages from the control."""
        self._message_tx_event += callback

    def remove_transmit_message_callback(self, callback):
        """Remove a callback for transmitted messages from the control."""
        self._message_tx_event -= callback

    def __init__(self, cntrl_type: str, control_id: str):
        """Control base type - all controls have these charactoristics and methods.

        Parameters
        ----------
        cntrl_type : str
            The type of control to implement
        control_id : str
            An unique control identity string. The control identity string must be a unique string for each control per device
        """
        # Dictionary to store CFG json
        self._cfg_columnar = []
        self._cfg_full_page = []
        self._cfg_full_page_no_columns = 0
        self.cntrl_type = cntrl_type.translate(BAD_CHARS)
        self.control_id = control_id.translate(BAD_CHARS)
        self._message_rx_event = Event()
        self._message_tx_event = Event()
        # This may break things but makes all controls able to be setup from tasks.
        self._message_rx_event += self._message_tx_event
        self._control_hdr_str = f"\t{{device_id}}\t{self.cntrl_type}\t{self.control_id}\t"

    def del_configs_columnar(self):
        """Deletes all the columnar config layout entries"""
        self._cfg_columnar = []

    def del_configs_full_page(self):
        """Deletes all the full page config layout entries"""
        self._cfg_columnar = []

    @property
    def state_str(self) -> str:
        """The current state of the control

        Returns
        -------
        str
            The current state of the control
        """
        return self._control_hdr_str

    @state_str.setter
    def state_str(self, val):
        self._message_tx_event(val)

    #  Use getter, setter properties to store the settings in the config dictionary
    @property
    def parent_id(self, index=0) -> str:
        """The parent control or deviceview this control belongs to

        Returns
        -------
        str
            The parent_id
        """
        return self._cfg_columnar[index]["parentID"]

    @parent_id.setter
    def parent_id(self, val: str, index=0):
        _val = val.translate(BAD_CHARS)
        self._cfg_columnar[index].parent_id = _val



import json
import unittest

from dashio import Selector


def _get_cfg_dict(cfg_list: list):
    json_str = cfg_list[0].rpartition('\t')[2]
    return json.loads(json_str)


class TestSelector(unittest.TestCase):

    def test_selector_control_type(self):
        test_control = Selector("SELECTORID")
        test_str_list = test_control._control_hdr_str.split('\t')
        self.assertEqual(test_str_list[2], 'SLCTR', "control type should be SLCTR")

    def test_selector_control_id(self):
        test_control = Selector("SELECTORID")
        test_str_list = test_control._control_hdr_str.split('\t')
        self.assertEqual(test_str_list[3], 'SELECTORID', "control type should be SELECTORID")


if __name__ == '__main__':
    unittest.main()



"""
MIT License

Copyright (c) 2020 DashIO-Connect

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
from ..constants import BAD_CHARS
from .control import Control, ControlPosition, ControlConfig, _get_color, _get_title_position, _get_direction_style, _get_precision
from .enums import Color, DirectionStyle, Precision, TitlePosition


class DirectionConfig(ControlConfig):
    """DirectionConfig"""
    def __init__(
        self,
        control_id: str,
        title: str,
        style: DirectionStyle,
        title_position: TitlePosition,
        pointer_color: Color,
        units: str,
        precision: Precision,
        calibration_angle: float,
        control_position: ControlPosition
    ) -> None:
        super().__init__(control_id, title, control_position, title_position)
        self.cfg["style"] = str(style.value)
        self.cfg["pointerColor"] = str(pointer_color.value)
        self.cfg["units"] = units
        self.cfg["precision"] = precision.value
        self.cfg["calAngle"] = calibration_angle

    @classmethod
    def from_dict(cls, cfg_dict: dict):
        """Instantiates DirectionConfig from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        DirectionConfig
        """
        tmp_cls = cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            _get_direction_style(cfg_dict["style"]),
            _get_title_position(cfg_dict["titlePosition"]),
            _get_color(cfg_dict["pointerColor"]),
            cfg_dict["units"],
            _get_precision(cfg_dict["precision"]),
            cfg_dict["calAngle"],
            ControlPosition(cfg_dict["xPositionRatio"], cfg_dict["yPositionRatio"], cfg_dict["widthRatio"], cfg_dict["heightRatio"])
        )
        tmp_cls.parent_id = cfg_dict["parentID"]
        return tmp_cls


class Direction(Control):
    """Direction control"""

    def add_config_columnar(self, config: DirectionConfig):
        if isinstance(config, DirectionConfig):
            config.cfg["calAngle"] = self.cal_angle
            config.cfg["ControlID"] = self.control_id
            self._cfg_columnar.append(config)

    def add_config_full_page(self, config: DirectionConfig):
        if isinstance(config, DirectionConfig):
            config.cfg["calAngle"] = self.cal_angle
            config.cfg["ControlID"] = self.control_id
            self._cfg_full_page.append(config)

    def __init__(
        self,
        control_id: str,
        title="A Control",
        style=DirectionStyle.DEG,
        title_position=TitlePosition.BOTTOM,
        pointer_color=Color.GREEN,
        units="",
        precision=Precision.OFF,
        calibration_angle=0,
        control_position=None
    ):
        """Direction Control

        Parameters
        ----------
        control_id : str
            An unique control identity string. The control identity string must be a unique string for each control per device
        title : str, optional
            Title of the control, by default "A Control"
        control_position : ControlPosition, optional
            The position of the control on a DeviceView, by default None
        title_position : TitlePosition, optional
            Position of the title when displayed on the iotdashboard app, by default None
        style : DirectionStyle, optional
            The Direction style to display, by default DirectionStyle.DEG
        pointer_color : Color, optional
            Color of the pointer, by default Color.GREEN
        units : str, optional
            Units to be displayed with the value, by default ""
        precision : Precision, optional
            Precision of the value displayed, by default Precision.OFF
        calibration_angle : int, optional
            Calibration angle offset, by default 0
        """
        super().__init__("DIR", control_id)
        self._cfg_columnar.append(
            DirectionConfig(
                control_id,
                title,
                style,
                title_position,
                pointer_color,
                units,
                precision,
                calibration_angle,
                control_position
            )
        )
        self._cal_angle = calibration_angle
        self._direction_value = 0
        self._direction_text = ""

    @property
    def cal_angle(self):
        """Retturns the calibration angle"""
        return self._cal_angle

    @classmethod
    def from_cfg_dict(cls, cfg_dict: dict):
        """Instatiates Direction from cfg dictionary

        Parameters
        ----------
        cfg_dict : dict
            A dictionary usually loaded from a config json from IoTDashboard App

        Returns
        -------
        Direction
        """
        tmp_cls = cls(
            cfg_dict["controlID"],
            cfg_dict["title"],
            _get_direction_style(cfg_dict["style"]),
            _get_title_position(cfg_dict["titlePosition"]),
            _get_color(cfg_dict["pointerColor"]),
            cfg_dict["units"],
            _get_precision(cfg_dict["precision"]),
            cfg_dict["calAngle"],
            ControlPosition(cfg_dict["xPositionRatio"], cfg_dict["yPositionRatio"], cfg_dict["widthRatio"], cfg_dict["heightRatio"])
        )
        tmp_cls.parent_id = cfg_dict["parentID"]
        return tmp_cls

    def get_state(self):
        if self._direction_text:
            s_str = self._control_hdr_str + f"{self._direction_value}\t{self._direction_text}\n"
        else:
            s_str = self._control_hdr_str + f"{self._direction_value}\n"
        return s_str

    @property
    def direction_value(self) -> float:
        """Direction value

        Returns
        -------
        float
            The direction
        """
        return self._direction_value

    @direction_value.setter
    def direction_value(self, val: float):
        self._direction_value = val
        if self._direction_text:
            s_str = self._control_hdr_str + f"{self._direction_value}\t{self._direction_text}\n"
        else:
            s_str = self._control_hdr_str + f"{self._direction_value}\n"
        self.state_str = s_str

    @property
    def direction_text(self) -> str:
        """Direction text

        Returns
        -------
        str
            Text to be displayed on the direction control
        """
        return self._direction_text

    @direction_text.setter
    def direction_text(self, val: str):
        self._direction_text = val.translate(BAD_CHARS)
        if self._direction_text:
            s_str = self._control_hdr_str + f"{self._direction_value}\t{self._direction_text}\n"
        else:
            s_str = self._control_hdr_str + f"{self._direction_value}\n"
        self.state_str = s_str
